from ._types import _BBoxSearch


def get_bbox_str(
    min_lat: str,
    min_lon: str,
    max_lat: str,
    max_lon: str,
):
    bounding_box = [min_lat, min_lon, max_lat, max_lon]
    return ",".join([str(float(x)) for x in bounding_box])


def parse_bbox_search(bounding_box: list[str | float | int] | None) -> _BBoxSearch:
    """Convert user's bounding box inputs to a query parameter, that can be used in search requests."""
    if bounding_box is not None:
        bbox = get_bbox_str(
            min_lat=str(float(bounding_box[1])),
            min_lon=str(float(bounding_box[0])),
            max_lat=str(float(bounding_box[3])),
            max_lon=str(float(bounding_box[2])),
        )
        return _BBoxSearch(bbox=bbox)
    return _BBoxSearch(bbox=None)



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    add_nadir_var,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_amcth(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens AM__CTH_2B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    nadir_idx = get_nadir_index(ds, nadir_idx=150)

    ds = ds.rename({"latitude": SWATH_LAT_VAR})
    ds = ds.rename({"longitude": SWATH_LON_VAR})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var=SWATH_LAT_VAR, swath_lon_var=SWATH_LON_VAR
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        time_var="time",
    )

    ds = rename_var_info(
        ds, "cloud_top_height_MSI", long_name="CTH from M-COP", units="m"
    )
    ds = rename_var_info(
        ds,
        "cloud_top_height_difference_ATLID_MSI",
        long_name="CTH difference (ATL $-$ MSI)",
        units="m",
    )

    ds = rename_var_info(
        ds,
        "quality_status",
        long_name="Quality status",
        units="",
    )

    ds["plot_cloud_top_height_difference_ATLID_MSI"] = ds[
        "cloud_top_height_difference_ATLID_MSI"
    ].copy()
    mask = ds["quality_status"].values == 4
    ds["plot_cloud_top_height_difference_ATLID_MSI"].values[mask] = 100e3

    ds = add_nadir_var(ds, "cloud_top_height_MSI")
    ds = add_nadir_var(ds, "cloud_top_height_difference_ATLID_MSI")
    ds = add_nadir_var(ds, "quality_status")
    ds = add_nadir_var(ds, "cloud_fraction")

    return ds



"""
**earthcarekit**

A Python package to simplify working with EarthCARE satellite data

See also:

- [Documentation](https://tropos-rsd.github.io/earthcarekit/)
- [Development status (GitHub)](https://github.com/TROPOS-RSD/earthcarekit)
- [License (Apache-2.0)](https://github.com/TROPOS-RSD/earthcarekit/blob/main/LICENSE)
- [Citation (Zenodo)](http://doi.org/10.5281/zenodo.16813294)

---

Copyright © 2025 TROPOS

---
"""

__author__ = "Leonard König"
__license__ = "Apache-2.0"
__version__ = "0.13.2"
__date__ = "2025-12-22"
__maintainer__ = "Leonard König"
__email__ = "koenig@tropos.de"
__title__ = "earthcarekit"

import sys

from .calval import *
from .download import ecdownload
from .plot import *
from .plot import FigureType, ecquicklook, ecswath
from .utils import ProfileData, geo, read
from .utils import statistics as stats
from .utils.config import (
    _warn_user_if_not_default_config_exists,
    create_example_config,
    get_config,
    get_default_config_filepath,
    set_config,
    set_config_maap_token,
    set_config_to_maap,
    set_config_to_oads,
)
from .utils.geo import geodesic, get_coord_between, get_coords, haversine
from .utils.ground_sites import GroundSite, get_ground_site
from .utils.logging import _setup_logging
from .utils.overpass import get_overpass_info
from .utils.read import *
from .utils.xarray_utils import (
    filter_index,
    filter_latitude,
    filter_radius,
    filter_time,
)

sys.modules[__name__ + ".geo"] = geo
sys.modules[__name__ + ".read"] = read
sys.modules[__name__ + ".stats"] = stats
__all__ = [
    "read",
    "stats",
    "geo",
    "ecquicklook",
    "ecswath",
    "ecdownload",
    "ProfileData",
    "filter_index",
    "filter_latitude",
    "filter_radius",
    "filter_time",
    "GroundSite",
    "get_ground_site",
    "get_overpass_info",
    "geodesic",
    "haversine",
    "get_coords",
    "get_coord_between",
    "get_config",
    "set_config",
    "set_config_maap_token",
    "set_config_to_maap",
    "set_config_to_oads",
    "create_example_config",
    "get_default_config_filepath",
    "FigureType",
]

_setup_logging()
_warn_user_if_not_default_config_exists()



import numpy as np
from matplotlib.axes import Axes
from numpy.typing import NDArray

from ....utils.geo import get_cumulative_distances
from ....utils.geo.string_formatting import format_latitude, format_longitude
from ....utils.time import to_timestamp
from ...figure.format_strings import format_float
from ...figure.ticks import add_ticks
from .style import AlongTrackAxisData, AlongTrackAxisStyle
from .time_ticks import format_time_ticks


def format_along_track_axis(
    ax: Axes,
    ax_style: AlongTrackAxisStyle,
    time: NDArray,
    tmin: np.datetime64,
    tmax: np.datetime64,
    tmin_original: np.datetime64,
    tmax_original: np.datetime64,
    lon: NDArray | None = None,
    lat: NDArray | None = None,
    num_ticks: int = 10,
) -> None:
    tmin = np.datetime64(to_timestamp(tmin))
    tmax = np.datetime64(to_timestamp(tmax))
    tmin_original = np.datetime64(to_timestamp(tmin_original))
    tmax_original = np.datetime64(to_timestamp(tmax_original))

    show_title = ax_style.title
    show_units = ax_style.units
    show_labels = ax_style.labels

    mask = np.logical_and(time >= tmin, time <= tmax)
    time = time[mask]
    if isinstance(lat, np.ndarray):
        lat = lat[mask]
    if isinstance(lon, np.ndarray):
        lon = lon[mask]

    if ax_style.data in [
        AlongTrackAxisData.TIME,
        AlongTrackAxisData.TIME_UTC,
        AlongTrackAxisData.TIME_LST,
    ]:
        show_title = False if show_title == False else True
        show_labels = False if show_labels == False else True

        show_utc = True
        show_lst = True
        if ax_style.data == AlongTrackAxisData.TIME_UTC:
            show_lst = False
        elif ax_style.data == AlongTrackAxisData.TIME_LST:
            show_utc = False

        format_time_ticks(
            ax=ax,
            time=time,
            tmin=tmin,
            tmax=tmax,
            lon=lon,
            show_title=show_title,
            show_utc=show_utc,
            show_lst=show_lst,
            show_tick_labels=show_labels,
            max_ticks=num_ticks,
        )
        return

    show_units = False if show_units == False else True
    show_labels = False if show_labels == False else True

    if (
        ax_style.data == AlongTrackAxisData.GEO
        and isinstance(lat, np.ndarray)
        and isinstance(lon, np.ndarray)
    ):
        show_title = True if show_title == True else False
        time_data = time
        lat_data = lat
        lon_data = lon
        if tmin < tmin_original:
            time_data = np.concatenate(([tmin], time_data))
            lat_data = np.concatenate(([np.nan], lat_data))
            lon_data = np.concatenate(([np.nan], lon_data))
        if tmax_original < tmax:
            time_data = np.concatenate((time_data, [tmax]))
            lat_data = np.concatenate((lat_data, [np.nan]))
            lon_data = np.concatenate((lon_data, [np.nan]))
        if show_title:
            if not show_units:
                title = r"Latitude/Logitude [$^\circ$N/$^\circ$E]"
            else:
                title = r"Latitude/Logitude"
        else:
            title = ""
        add_ticks(
            ax=ax,
            ax_data=time_data,
            major_tick_count=num_ticks,
            fontsize="small",
            title=title,
            show_tick_labels=show_labels,
            tick_data=lat_data,
            format_function=format_latitude if show_units else None,
            tick_data2=lon_data,
            format_function2=format_longitude if show_units else None,
        )
    elif ax_style.data == AlongTrackAxisData.LAT and isinstance(lat, np.ndarray):
        show_title = True if show_title == True else False
        time_data = time
        lat_data = lat
        if tmin < tmin_original:
            time_data = np.concatenate(([tmin], time_data))
            lat_data = np.concatenate(([np.nan], lat_data))
        if tmax_original < tmax:
            time_data = np.concatenate((time_data, [tmax]))
            lat_data = np.concatenate((lat_data, [np.nan]))
        if show_title:
            if not show_units:
                title = r"Latitude [$^\circ$N]"
            else:
                title = r"Latitude"
        else:
            title = ""
        add_ticks(
            ax=ax,
            ax_data=time_data,
            major_tick_count=num_ticks,
            fontsize="small",
            title=title,
            show_tick_labels=show_labels,
            tick_data=lat_data,
            format_function=format_latitude if show_units else None,
        )
    elif ax_style.data == AlongTrackAxisData.LON and isinstance(lon, np.ndarray):
        show_title = True if show_title == True else False
        time_data = time
        lon_data = lon
        if tmin < tmin_original:
            time_data = np.concatenate(([tmin], time_data))
            lon_data = np.concatenate(([np.nan], lon_data))
        if tmax_original < tmax:
            time_data = np.concatenate((time_data, [tmax]))
            lon_data = np.concatenate((lon_data, [np.nan]))
        if show_title:
            if not show_units:
                title = r"Logitude [$^\circ$E]"
            else:
                title = r"Logitude"
        else:
            title = ""
        add_ticks(
            ax=ax,
            ax_data=time_data,
            major_tick_count=num_ticks,
            fontsize="small",
            title=title,
            show_tick_labels=show_labels,
            tick_data=lon_data,
            format_function=format_longitude if show_units else None,
        )
    elif (
        ax_style.data == AlongTrackAxisData.DISTANCE
        and isinstance(lat, np.ndarray)
        and isinstance(lon, np.ndarray)
    ):
        show_title = False if show_title == False else True
        distances = get_cumulative_distances(lat, lon, units="km")
        if show_title:
            title = r"Distance [km]"
        else:
            title = ""
        add_ticks(
            ax=ax,
            ax_data=time,
            major_tick_count=num_ticks,
            fontsize="small",
            title=title,
            tick_data=distances,
            format_function=format_float,
            show_tick_labels=show_labels,
            in_linspace=False,
        )
    elif ax_style.data == AlongTrackAxisData.COUNT:
        show_title = False if show_title == False else True
        if show_title:
            title = r"Samples"
        else:
            title = ""
        add_ticks(
            ax=ax,
            ax_data=time,
            major_tick_count=num_ticks,
            fontsize="small",
            title=title,
            tick_data=np.arange(len(time)),
            format_function=format_float,
            show_tick_labels=show_labels,
            in_linspace=False,
        )
    else:
        ax.set_xticks([], [])



import warnings
from typing import Literal

import numpy as np
import xarray as xr
from scipy.interpolate import griddata  # type: ignore

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    ELEVATION_VAR,
    HEIGHT_VAR,
    VERTICAL_DIM,
)
from ....rolling_mean import rolling_mean_2d
from ....statistics import nan_mean
from ....xarray_utils import filter_time, merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def get_depol_profile(
    ds: xr.Dataset,
    cpol_cleaned_var: str = "cpol_cleaned_for_ratio_calculation",
    xpol_cleaned_var: str = "xpol_cleaned_for_ratio_calculation",
):
    cpol = ds[cpol_cleaned_var].data
    xpol = ds[xpol_cleaned_var].data
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mean_xpol_bsc = np.nanmean(xpol, axis=0)
        mean_mie_bsc = np.nanmean(cpol, axis=0)
    return mean_xpol_bsc / mean_mie_bsc


def add_scattering_ratio(
    ds_anom: xr.Dataset,
    formula: Literal["x/c", "(c+x)/r", "(c+x+r)/r"],
    rolling_w: int = 20,
    near_zero_tolerance: float = 2e-7,
    smooth: bool = True,
    skip_height_above_elevation: int = 300,
    cpol_var: str = "mie_attenuated_backscatter",
    xpol_var: str = "crosspolar_attenuated_backscatter",
    ray_var: str = "rayleigh_attenuated_backscatter",
    elevation_var: str = ELEVATION_VAR,
    height_var: str = HEIGHT_VAR,
    height_dim: str = VERTICAL_DIM,
) -> xr.Dataset:
    """
    Compute scattering ratio from attenuated backscatter signals given a formula: "x/c", "(c+x)/r", or "(c+x+r)/r".

    This function derives the scattering ratio from cross-polarized (`XPOL`), co-polarized (`CPOL`) and rayleigh (`RAY`) attenuated backscatter signals.
    Signals below the surface are masked, by default with a vertical margin on 300 meters above elevation to remove potential surface return.
    Also, signals are smoothed (or "cleaned") with a rolling mean, and near-zero divisions are suppressed and set to NaN instead.
    In the resulting dataset, the ratio curtain and a ratio profile calculated from mean profiles of the full dataset (e.g., mean(`XPOL`)/mean(`CPOL`)).

    Args:
        ds_anom (xr.Dataset): ATL_NOM_1B dataset containing the attenuated backscatter signals.
        formula (Literal["x/c", "(c+x)/r", "(c+x+r)/r"]): Formula used to calculate the scattering ratio.
        rolling_w (int, optional): Window size for rolling mean smoothing. Defaults to 20.
        near_zero_tolerance (float, optional): Tolerance for masking near-zero denominators. Defaults to 2e-7.
        smooth (bool, optional): Whether to apply rolling mean smoothing. Defaults to True.
        skip_height_above_elevation (int, optional): Vertical margin above surface elevation to mask in meters. Defaults to 300.
        cpol_var (str, optional): Input co-polar variable name. Defaults to "mie_attenuated_backscatter".
        xpol_var (str, optional): Input cross-polar variable name. Defaults to "crosspolar_attenuated_backscatter".
        ray_var (str, optional): Input rayleigh variable name. Defaults to "rayleigh_attenuated_backscatter".
        elevation_var (str, optional): Elevation variable name. Defaults to ELEVATION_VAR.
        height_var (str, optional): Height variable name. Defaults to HEIGHT_VAR.
        height_dim (str, optional): Height dimension name. Defaults to VERTICAL_DIM.

    Returns:
        xr.Dataset: xr.Dataset: Dataset with added ratio curtain and ratio profile from mean profiles.
    """

    if formula.lower() not in ["x/c", "(c+x)/r", "(c+x+r)/r"]:
        raise ValueError(
            f"invalid formula '{formula}', expected 'x/c', '(c+x)/r' or '(c+x+r)/r'"
        )

    cpol_cleaned_var: str = "cpol_cleaned_for_ratio_calculation"
    xpol_cleaned_var: str = "xpol_cleaned_for_ratio_calculation"
    ray_cleaned_var: str = "ray_cleaned_for_ratio_calculation"

    cpol_da = ds_anom[cpol_var].copy()
    xpol_da = ds_anom[xpol_var].copy()
    if formula == "x/c":
        ray_da = xpol_da
    else:
        ray_da = ds_anom[ray_var].copy()

    def _calc(c, x, r):
        if formula == "x/c":
            return x / c
        elif formula == "(c+x)/r":
            return (c + x) / r
        elif formula == "(c+x+r)/r":
            return (c + x + r) / r

    def _get_near_zero_mask(c, x, r):
        if formula == "x/c":
            return np.isclose(c, 0, atol=near_zero_tolerance)
        elif formula == "(c+x)/r":
            return np.isclose(r, 0, atol=near_zero_tolerance)
        elif formula == "(c+x+r)/r":
            return np.isclose(r, 0, atol=near_zero_tolerance)

    def _get_long_name():
        if formula == "x/c":
            return "Depol. ratio from cross- and co-polar atten. part. bsc."
        elif formula == "(c+x)/r":
            return "Total part. to ray. atten. bsc. ratio"
        elif formula == "(c+x+r)/r":
            return "Total to ray. atten. bsc. ratio"

    def _get_ratio_var():
        if formula == "x/c":
            return "depol_ratio"
        elif formula == "(c+x)/r":
            return "cpol_xpol_to_ray_ratio"
        elif formula == "(c+x+r)/r":
            return "cpol_xpol_ray_to_ray_ratio"

    ratio_var = _get_ratio_var()
    ratio_from_means_var = f"{ratio_var}_from_means"

    ds_anom[ratio_var] = _calc(cpol_da, xpol_da, ray_da)
    rename_var_info(
        ds_anom,
        ratio_var,
        name=ratio_var,
        long_name=_get_long_name(),
        units="",
    )

    elevation = (
        ds_anom[elevation_var].data.copy()[:, np.newaxis] + skip_height_above_elevation
    )
    mask_surface = ds_anom[height_var].data[0].copy() < elevation

    cpol = ds_anom[cpol_var].data
    xpol = ds_anom[xpol_var].data
    if formula == "x/c":
        ray = xpol
    else:
        ray = ds_anom[ray_var].data

    cpol[mask_surface] = np.nan
    xpol[mask_surface] = np.nan
    ray[mask_surface] = np.nan

    if smooth:
        cpol = rolling_mean_2d(cpol, rolling_w, axis=0)
        xpol = rolling_mean_2d(xpol, rolling_w, axis=0)
        ray = rolling_mean_2d(ray, rolling_w, axis=0)

    ds_anom[ratio_var].data = _calc(cpol, xpol, ray)
    ds_anom[ratio_var] = ds_anom[ratio_var].assign_attrs(
        {
            "earthcarekit": "Added by earthcarekit: Intended for use in curtain plots only!",
        }
    )

    if smooth:
        near_zero_mask = _get_near_zero_mask(cpol, xpol, ray)
        ds_anom[ratio_var].data[near_zero_mask] = np.nan
        cpol[near_zero_mask] = np.nan
        xpol[near_zero_mask] = np.nan
        ray[near_zero_mask] = np.nan

    ds_anom[xpol_cleaned_var] = ds_anom[xpol_var].copy()
    ds_anom[xpol_cleaned_var].data = xpol
    ds_anom[xpol_cleaned_var] = ds_anom[xpol_cleaned_var].assign_attrs(
        {
            "earthcarekit": f"Added by earthcarekit: Rolling mean applied (w={rolling_w}) and near-zero values removed (tolerance={near_zero_tolerance})"
        }
    )

    ds_anom[cpol_cleaned_var] = ds_anom[cpol_var].copy()
    ds_anom[cpol_cleaned_var].data = cpol
    ds_anom[cpol_cleaned_var] = ds_anom[cpol_cleaned_var].assign_attrs(
        {
            "earthcarekit": f"Added by earthcarekit: Rolling mean applied (w={rolling_w}) and near-zero values removed (tolerance={near_zero_tolerance})"
        }
    )

    if formula == "x/c":
        ds_anom[ray_cleaned_var] = ds_anom[ray_var].copy()
        ds_anom[ray_cleaned_var].data = ray
        ds_anom[ray_cleaned_var] = ds_anom[ray_cleaned_var].assign_attrs(
            {
                "earthcarekit": f"Added by earthcarekit: Rolling mean applied (w={rolling_w}) and near-zero values removed (tolerance={near_zero_tolerance})"
            }
        )

    ratio_mean = _calc(
        nan_mean(cpol, axis=0),
        nan_mean(xpol, axis=0),
        nan_mean(ray, axis=0),
    )

    ds_anom[ratio_from_means_var] = xr.DataArray(
        data=ratio_mean,
        dims=[height_dim],
        attrs={
            "long_name": _get_long_name(),
            "units": "",
            "earthcarekit": "Added by earthcarekit: Scattering ratio profile calculated from the mean profiles",
        },
    )

    return ds_anom


def add_depol_ratio(
    ds_anom: xr.Dataset,
    rolling_w: int = 20,
    near_zero_tolerance: float = 2e-7,
    smooth: bool = True,
    skip_height_above_elevation: int = 300,
    cpol_var: str = "mie_attenuated_backscatter",
    xpol_var: str = "crosspolar_attenuated_backscatter",
    elevation_var: str = ELEVATION_VAR,
    height_var: str = HEIGHT_VAR,
    height_dim: str = VERTICAL_DIM,
) -> xr.Dataset:
    """
    Compute depolarization ratio (`DPOL` = `XPOL`/`CPOL`) from attenuated backscatter signals.

    This function derives the depol. ratio from cross-polarized (`XPOL`) and co-polarized (`CPOL`) attenuated backscatter signals.
    Signals below the surface are masked, by default with a vertical margin on 300 meters above elevation to remove potential surface return.
    Also, signals are smoothed (or "cleaned") with a rolling mean, and near-zero divisions are suppressed and set to NaN instead.
    In the resulting dataset, the ratio curtain and a ratio profile calculated from mean profiles of the full dataset (e.g., mean(`XPOL`)/mean(`CPOL`)).

    Args:
        ds_anom (xr.Dataset): ATL_NOM_1B dataset containing cross- and co-polar attenuated backscatter.
        rolling_w (int, optional): Window size for rolling mean smoothing. Defaults to 20.
        near_zero_tolerance (float, optional): Tolerance for masking near-zero `CPOL` (i.e., denominators). Defaults to 2e-7.
        smooth (bool, optional): Whether to apply rolling mean smoothing. Defaults to True.
        skip_height_above_elevation (int, optional): Vertical margin above surface elevation to mask in meters. Defaults to 300.
        cpol_var (str, optional): Input co-polar variable name. Defaults to "mie_attenuated_backscatter".
        xpol_var (str, optional): Input cross-polar variable name. Defaults to "crosspolar_attenuated_backscatter".
        elevation_var (str, optional): Elevation variable name. Defaults to ELEVATION_VAR.
        height_var (str, optional): Height variable name. Defaults to HEIGHT_VAR.
        height_dim (str, optional): Height dimension name. Defaults to VERTICAL_DIM.

    Returns:
        xr.Dataset: Dataset with added depol. ratio, cleaned signals, and depol. ratio profile from mean profiles.
    """
    return add_scattering_ratio(
        ds_anom=ds_anom,
        formula="x/c",
        rolling_w=rolling_w,
        near_zero_tolerance=near_zero_tolerance,
        smooth=smooth,
        skip_height_above_elevation=skip_height_above_elevation,
        cpol_var=cpol_var,
        xpol_var=xpol_var,
        elevation_var=elevation_var,
        height_var=height_var,
        height_dim=height_dim,
    )


def read_product_anom(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens ATL_NOM_1B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds["original_time"] = ds["time"].copy()
    ds["original_time"] = ds["original_time"].assign_attrs(
        {"earthcarekit": "Added by earthcarekit: A copy of the original time variable."}
    )
    ds["time"].data = ds["time"].data + np.timedelta64(-2989554432, "ns")
    ds["time"] = ds["time"].assign_attrs(
        {
            "earthcarekit": 'Modified by earthcarekit: Since ATLID is angled backwards a time shift of around 3 seconds (here deltatime=-2989554432 ns) is applied to facilitate plotting with L2 products. The original time is stored in the variable "original_time".'
        }
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="height",
        track_lat_var="ellipsoid_latitude",
        track_lon_var="ellipsoid_longitude",
        height_var="sample_altitude",
        time_var="time",
        temperature_var="layer_temperature",
        elevation_var="surface_elevation",
        land_flag_var="land_flag",
    )
    ds = rename_var_info(
        ds,
        "mie_attenuated_backscatter",
        "Co-polar atten. part. bsc.",
        "Co-polar atten. part. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = rename_var_info(
        ds,
        "rayleigh_attenuated_backscatter",
        "Ray. atten. bsc.",
        "Ray. atten. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = rename_var_info(
        ds,
        "crosspolar_attenuated_backscatter",
        "Cross-polar atten. part. bsc.",
        "Cross-polar atten. part. bsc.",
        "m$^{-1}$ sr$^{-1}$",
    )
    ds = add_depol_ratio(ds)

    return ds



from typing import Final

CM_AS_INCH: Final[float] = 1.0 / 2.54
FIGURE_HEIGHT_CURTAIN: Final[float] = 8.0 * CM_AS_INCH
FIGURE_WIDTH_CURTAIN: Final[float] = FIGURE_HEIGHT_CURTAIN * 3.0
FIGURE_HEIGHT_SWATH: Final[float] = FIGURE_HEIGHT_CURTAIN * 0.5
FIGURE_WIDTH_SWATH: Final[float] = FIGURE_WIDTH_CURTAIN
FIGURE_HEIGHT_PROFILE: Final[float] = FIGURE_HEIGHT_CURTAIN
FIGURE_WIDTH_PROFILE: Final[float] = FIGURE_HEIGHT_CURTAIN * 0.75
FIGURE_MAP_HEIGHT: Final[float] = FIGURE_HEIGHT_CURTAIN
FIGURE_MAP_WIDTH: Final[float] = FIGURE_HEIGHT_CURTAIN
FIGURE_HEIGHT_LINE: Final[float] = FIGURE_HEIGHT_CURTAIN * 0.5
FIGURE_WIDTH_LINE: Final[float] = FIGURE_WIDTH_CURTAIN



import re
from dataclasses import dataclass
from typing import Sequence, Tuple, TypeAlias, Union

import matplotlib.colors as mcolors
import numpy as np

from .color_contrast import get_best_bw_contrast_color

Numeric: TypeAlias = int | float
ColorLike: TypeAlias = str | Sequence[Numeric]

_custom_colors = {
    "ec:darkblue": "#003F5B",
    "ec:blue": "#009ADE",
    "ec:lightblue": "#74BBD1",
    "ec:darkpurple": "#671B85",
    "ec:purple": "#AF58BA",
    "ec:lightpurple": "#CEABDB",
    "ec:darkred": "#A01154",
    "ec:red": "#FF1F5B",
    "ec:lightred": "#DB949E",
    "ec:darkorange": "#A85C16",
    "ec:orange": "#F28522",
    "ec:lightorange": "#E4BE6E",
    "ec:darkyellow": "#968C00",
    "ec:yellow": "#FFC61E",
    "ec:lightyellow": "#F3E490",
    "ec:darkgreen": "#008F7C",
    "ec:green": "#00CD6C",
    "ec:lightgreen": "#6CDFB9",
    "ec:grey": "#A0B1BA",
    "ec:gray": "#A0B1BA",
    "ec:brown": "#A6761D",
    "ec:olive": "#BCBD22",
    "ec:crimson": "#E9002D",
    "ec:amber": "#FFAA00",
    "ec:lime": "#00B000",
    "ec:berry": "#C40F5B",
    "ec:carrot": "#FD8D3C",
    "ec:petrol": "#089099",
    "ec:black": "#000000",
    "ec:white": "#FFFFFF",
    "ec:tropopause": "#ff13f0",
    "ec:earthcare": "#FF1F5B",
    "ec:polly": "#009ADE",
    "ec:cloudnet": "#AF58BA",
    "ec:elevation": "#856b59",
    "ec:land": "#856B59",
    "ec:water": "#74BBD1",
}


@dataclass(frozen=True)
class Color(str):
    """Represents a color with convenient conversion, blending, and alpha support.

    Extends `str` to store a color as a hex string while providing methods
    to access RGB/RGBA, set transparency, blend with other colors, and
    normalize input from various formats.

    Attributes:
        input (ColorLike): Original input used to create the color.
        name (str | None): Optional name of the color.
        is_normalized (bool): Whether the color values are normalized (0-1).
    """

    input: ColorLike
    name: str | None = None
    is_normalized: bool = False

    def __new__(
        cls,
        color_input: "Color" | ColorLike,
        name: str | None = None,
        is_normalized: bool = False,
    ):
        """Create a Color instance from a color-like input."""
        hex_color = cls._to_hex(color_input, is_normalized=is_normalized)
        return str.__new__(cls, hex_color)

    def __init__(
        self,
        color_input: "Color" | ColorLike,
        name: str | None = None,
        is_normalized: bool = False,
    ):
        """Initialize Color attributes."""
        object.__setattr__(self, "input", color_input)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "is_normalized", is_normalized)

    def __hash__(self):
        """Return the hash of the color string."""
        return hash(str(self))

    @classmethod
    def _rgb_str_to_hex(
        cls,
        rgb_string: str,
        is_normalized: bool = False,
    ) -> str:
        """Convert an 'rgb(...)' string to a hex color."""
        if (not is_normalized and re.match(r"^rgb\(\d*,\d*,\d*\)$", rgb_string)) or (
            is_normalized
            and re.match(r"^rgb\(\d*\.?\d*,\d*\.?\d*,\d*\.?\d*\)$", rgb_string)
        ):
            rgb_str_list = rgb_string[4:-1].split(",")
            if is_normalized:
                rgb_tuple = tuple(float(v) for v in rgb_str_list)
            else:
                rgb_tuple = tuple(int(v) for v in rgb_str_list)
            return cls._rgb_tuple_to_hex(rgb_tuple, is_normalized=is_normalized)
        raise ValueError(f"Invalid rgb color: '{rgb_string}'")

    @classmethod
    def _rgba_str_to_hex(
        cls,
        rgba_string: str,
        is_normalized: bool = False,
    ) -> str:
        """Convert an 'rgba(...)' string to a hex color."""
        if (
            not is_normalized
            and re.match(r"^rgba\(\d*,\d*,\d*,\d*\.?\d*\)$", rgba_string)
        ) or (
            is_normalized
            and re.match(
                r"^rgba\(\d*\.?\d*,\d*\.?\d*,\d*\.?\d*,\d*\.?\d*\)$", rgba_string
            )
        ):
            rgba_str_list = rgba_string[5:-1].split(",")
            if float(rgba_str_list[-1]) > 1:
                raise ValueError(
                    f"Invalid alpha value (must be float between 0 and 1): '{rgba_string}'"
                )
            if is_normalized:
                rgba_tuple = tuple(float(v) for v in rgba_str_list)
            else:
                rgba_tuple = tuple(
                    int(v) if i < 3 else float(v) for i, v in enumerate(rgba_str_list)
                )
            return cls._rgba_tuple_to_hex(rgba_tuple, is_normalized=is_normalized)
        raise ValueError(f"Invalid rgba color: '{rgba_string}'")

    @classmethod
    def _rgb_tuple_to_hex(
        cls,
        rgb_tuple: Tuple[int, ...] | Tuple[float, ...],
        is_normalized: bool = False,
    ) -> str:
        """Convert an RGB tuple to a hex color string."""
        if is_normalized:
            rgb_tuple = tuple(int(v * 255) for v in rgb_tuple)
        is_all_int = all(isinstance(v, int | np.integer) for v in rgb_tuple)
        is_all_in_range = all(0 <= v <= 255 for v in rgb_tuple)
        if is_all_int and is_all_in_range:
            return "#{:02X}{:02X}{:02X}".format(*rgb_tuple)
        raise ValueError(f"Invalid rgb tuple: '{rgb_tuple}'")

    @classmethod
    def _rgba_tuple_to_hex(
        cls,
        rgba_tuple: Tuple[int | float, ...],
        is_normalized: bool = False,
    ) -> str:
        """Convert an RGBA tuple to a hex color string."""
        if is_normalized:
            rgba_tuple = tuple(
                int(v * 255) if i < 3 else v for i, v in enumerate(rgba_tuple)
            )
        is_all_int = all(
            isinstance(v, int | np.integer | float | np.floating) for v in rgba_tuple
        )
        is_all_in_range = all(
            0 <= v <= 255 if i < 3 else 0 <= v <= 1 for i, v in enumerate(rgba_tuple)
        )
        if is_all_int and is_all_in_range:
            rgba_tuple = tuple(
                int(v) if i < 3 else float(v) for i, v in enumerate(rgba_tuple)
            )
            rgba_int_tuple = tuple(
                int(v) if i < 3 else int(float(v) * 255)
                for i, v in enumerate(rgba_tuple)
            )
            return "#{:02X}{:02X}{:02X}{:02X}".format(*rgba_int_tuple)
        raise ValueError(f"Invalid rgba tuple: '{rgba_tuple}'")

    @classmethod
    def _hex_str_to_hex(
        cls,
        hex_string: str,
    ) -> str:
        """Normalize a hex string to standard 6- or 8-character format."""
        c = hex_string.upper()
        if re.match(r"^#[A-F0-9]{3}$", c):
            c = f"#{c[1]*2}{c[2]*2}{c[3]*2}"
        elif re.match(r"^#[A-F0-9]{4}$", c):
            c = f"#{c[1]*2}{c[2]*2}{c[3]*2}{c[4]*2}"
        if not re.match(r"^#[A-F0-9]{6}$", c) and not re.match(r"^#[A-F0-9]{8}$", c):
            raise ValueError(f"Invalid hex color: '{hex_string}'")
        return c

    @classmethod
    def _to_hex(
        cls,
        color: str | Sequence,
        is_normalized: bool = False,
    ) -> str:
        """Convert a color input of various formats to a hex string."""
        if isinstance(color, str):
            c_str = color.strip().replace(" ", "").lower()
            if c_str.startswith("#"):
                return cls._hex_str_to_hex(c_str)
            elif c_str.startswith("rgb("):
                return cls._rgb_str_to_hex(c_str)
            elif c_str.startswith("rgba("):
                return cls._rgba_str_to_hex(c_str)
            elif c_str.startswith("rgb255("):
                return cls._rgb_str_to_hex(c_str.replace("rgb255(", "rgb("))
            elif c_str.startswith("rgba255("):
                return cls._rgba_str_to_hex(c_str.replace("rgba255(", "rgba("))
            elif c_str.startswith("rgb01("):
                return cls._rgb_str_to_hex(
                    c_str.replace("rgb01(", "rgb("), is_normalized=True
                )
            elif c_str.startswith("rgba01("):
                return cls._rgba_str_to_hex(
                    c_str.replace("rgba01(", "rgba("), is_normalized=True
                )
            else:
                try:
                    return _custom_colors[color].upper()
                except KeyError as e:
                    pass
                return mcolors.to_hex(color).upper()
        elif isinstance(color, (Sequence, np.ndarray)):
            if len(color) > 0:
                if isinstance(color[0], float):
                    is_normalized = True
                else:
                    is_normalized = False
            c_tup = tuple(float(v) for v in color)
            if len(c_tup) == 3:
                if is_normalized:
                    c_tup = tuple(float(v) for v in color)
                else:
                    c_tup = tuple(int(v) for v in color)
                return cls._rgb_tuple_to_hex(c_tup, is_normalized=is_normalized)
            elif len(c_tup) == 4:
                return cls._rgba_tuple_to_hex(c_tup, is_normalized=is_normalized)
        raise TypeError(f"Invalid type for input color ({type(color)}: {color})")

    @property
    def hex(self) -> str:
        """Returns the hex color string."""
        return str(self).upper()

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Returns the RGB tuple with values in the 0-255 range."""
        hex_str = self.lstrip("#")
        return (
            int(hex_str[0:2], 16),
            int(hex_str[2:4], 16),
            int(hex_str[4:6], 16),
        )

    @property
    def alpha(self) -> float:
        """Returns the transparency alpha value in the 0-1 range."""
        if len(self) == 9:
            return int(str(self).upper()[7::], 16) / 255
        return 1.0

    @property
    def rgba(self) -> Tuple[float, float, float, float]:
        """Returns the RGBA tuple with values in the 0-1 range."""
        hex_str = self.lstrip("#")
        return (
            int(hex_str[0:2], 16) / 255,
            int(hex_str[2:4], 16) / 255,
            int(hex_str[4:6], 16) / 255,
            self.alpha,
        )

    def set_alpha(self, value: float) -> "Color":
        """Returns the same color with the given transparency alpha value applied."""
        if not 0 <= value <= 1:
            raise ValueError(
                f"Invalid alpha value: '{value}' (must be in the 0-1 range)"
            )
        return Color(self.hex[0:7] + "{:02X}".format(int(value * 255)), name=self.name)

    def blend(
        self, value: float, blend_color: "Color" | ColorLike = "white"
    ) -> "Color":
        """Returns the same color blended with a second color."""
        original_color = self.rgb
        blend_color = Color(blend_color).rgb
        new_color = Color(
            tuple(
                int(round((1 - value) * bc + value * oc))
                for oc, bc in zip(original_color, blend_color)
            )
        )
        return new_color

    @classmethod
    def from_optional(
        cls,
        color_input: "Color" | ColorLike | None,
        alpha: float | None = None,
    ) -> Union["Color", None]:
        """Parses optional color input and returns a `Color` instance or `None`."""
        if color_input is None:
            return None
        elif isinstance(alpha, float):
            return cls(color_input).set_alpha(alpha)
        else:
            return cls(color_input)

    def is_close_to_white(self, threshold: float = 0.1) -> bool:
        """Check if the color is close to white."""
        rgb01 = 1 - (np.array(self.rgb) / 255)
        return bool(np.all(rgb01 < threshold))

    def get_best_bw_contrast_color(self) -> "Color":
        """
        Return black or white color depending on best contrast according to WCAG 2.0.

        See https://www.w3.org/TR/WCAG20/
        """
        return Color(get_best_bw_contrast_color(self.rgb))



import numpy as np
import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
    UNITS_KELVIN,
    UNITS_MSI_RADIANCE,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    drop_samples_with_missing_geo_data_along_track,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def _get_bitmasks(ds: xr.Dataset, var: str, n: int):
    bits = ds[var].values.astype(np.uint16)

    masks = {i: 1 << i for i in range(n + 1)}
    bitmasks = {name: (bits & mask) > 0 for name, mask in masks.items()}

    return bits, bitmasks


def _get_dominant_classes(ds: xr.Dataset, var: str, n: int):
    bits, bitmasks = _get_bitmasks(ds=ds, var=var, n=n)

    new_values = np.zeros(bits.shape)
    for i in range(1, n + 1):
        new_values[bitmasks[i]] = i

    return new_values


def add_surface_classification_plot_var(
    ds: xr.Dataset,
    var: str = "surface_classification",
    n: int = 8,
):
    """Adds a plottable variable for the M-CM "surface_classification" to the given dataset, called "plot_surface_classification"."""
    new_values = _get_dominant_classes(ds, var=var, n=n)

    new_var = f"plot_{var}"
    ds[new_var] = ds[var].copy()
    ds[new_var].values = new_values
    ds[new_var] = ds[new_var].assign_attrs(
        {
            "long_name": "Surface classification",
            "definition": "0: Undefined, 1: Water, 2: Land, 3: Desert, 4: Vegetation NDVI, 5: Snow XMET, 6: Snow NDSI, 7: Sea ice XMET, 8: Sunglint",
            "units": "",
            "earthcarekit": "Added by earthcarekit: class integers converted from bitwise",
        }
    )

    return ds


def add_quality_status_plot_var(ds, var: str, n: int = 4):
    """Adds a plottable variable for the M-CM "cloud_[mask/type/phase]_quality_status" variable to the given dataset, called "plot_cloud_[mask/type/phase]_quality_status"."""
    if var not in [
        "cloud_mask_quality_status",
        "cloud_type_quality_status",
        "cloud_phase_quality_status",
    ]:
        raise ValueError(f"invalid MCM quality status variable '{var}'")

    new_values = _get_dominant_classes(ds, var=var, n=n)

    new_var = f"plot_{var}"
    ds[new_var] = ds[var].copy()
    ds[new_var].values = new_values
    ds[new_var] = ds[new_var].assign_attrs(
        {
            "definition": "0: Undefined, 1: Poor, 2: Low, 3: Medium, 4: High",
            "units": "",
            "earthcarekit": "Added by earthcarekit: class integers converted from bitwise",
        }
    )

    return ds


def read_product_mcm(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens MSI_CM__2A file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = drop_samples_with_missing_geo_data_along_track(
        ds=ds,
        swath_lat_var="latitude",
        along_track_dim="along_track",
        across_track_dim="across_track",
    )

    nadir_idx = get_nadir_index(ds, nadir_idx=266)
    ds = ds.rename({"latitude": SWATH_LAT_VAR})
    ds = ds.rename({"longitude": SWATH_LON_VAR})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var=SWATH_LAT_VAR, swath_lon_var=SWATH_LON_VAR
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        time_var="time",
    )

    ds = add_surface_classification_plot_var(ds)
    ds = add_quality_status_plot_var(ds, var="cloud_mask_quality_status")
    ds = add_quality_status_plot_var(ds, var="cloud_type_quality_status")
    ds = add_quality_status_plot_var(ds, var="cloud_phase_quality_status")

    return ds



from logging import Logger
from typing import Literal, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import (
    ACROSS_TRACK_DIM,
    ALONG_TRACK_DIM,
    CM_AS_INCH,
    DEFAULT_PROFILE_SHOW_STEPS,
    TIME_VAR,
    VERTICAL_DIM,
)
from ....utils.time import TimedeltaLike, TimeRangeLike
from ....utils.typing import DistanceRangeLike
from ....utils.xarray_utils import filter_radius, filter_time
from ...figure import (
    CurtainFigure,
    ECKFigure,
    FigureType,
    LineFigure,
    MapFigure,
    ProfileFigure,
    create_multi_figure_layout,
)
from .._cli import print_progress
from .._quicklook_results import QuicklookFigure
from ..set_default_height_range import set_none_height_range_to_default


def is_curtain_var(ds: xr.Dataset, var: str):
    return all(s in ds[var].dims for s in (ALONG_TRACK_DIM, VERTICAL_DIM))


def is_swath_var(ds: xr.Dataset, var: str):
    return all(s in ds[var].dims for s in (ALONG_TRACK_DIM, ACROSS_TRACK_DIM))


def is_1d_var(ds: xr.Dataset, var: str):
    return ds[var].dims == (ALONG_TRACK_DIM,)


def ecquicklook_actc(
    ds: xr.Dataset,
    vars: list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = False,
    site: str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = None,
    ds_tropopause: xr.Dataset | None = None,
    ds_elevation: xr.Dataset | None = None,
    ds_temperature: xr.Dataset | None = None,
    resolution: Literal["low", "medium", "high", "l", "m", "h"] = "medium",
    closest_profile: bool = True,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    show_steps: bool = DEFAULT_PROFILE_SHOW_STEPS,
    mode: Literal["fast", "exact"] = "fast",
) -> QuicklookFigure:

    map_figs: list[ECKFigure] = []

    common_kwargs = dict(
        time_range=time_range,
        site=site,
        radius_km=radius_km,
        selection_max_time_margin=selection_max_time_margin,
    )

    height_range = set_none_height_range_to_default(height_range, -250, 20e3)

    show_profile = False

    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    map_rows = None
    if show_maps:
        map_rows = [
            FigureType.MAP_1_ROW,
            FigureType.MAP_2_ROW,
        ]
    layout = create_multi_figure_layout(
        rows=[
            FigureType.CURTAIN,
            FigureType.CURTAIN,
            FigureType.CURTAIN,
        ],
        map_rows=map_rows,
    )

    if show_maps:
        ax_map1 = layout.axs_map[0]
        ax_map2 = layout.axs_map[1]

        fig_map1 = MapFigure(ax=ax_map1)
        fig_map1 = fig_map1.ecplot(
            ds=ds,
            view="global",
            **common_kwargs,  # type: ignore
        )
        fig_map2 = MapFigure(
            ax=ax_map2,
            style="blue_marble",
            coastlines_resolution="50m",
            show_right_labels=False,
            show_top_labels=False,
        )
        fig_map2 = fig_map2.ecplot(
            ds=ds,
            view="overpass",
            **common_kwargs,  # type: ignore
        )

        map_figs = [fig_map1, fig_map2]

    ax1 = layout.axs[0]
    ax2 = layout.axs[1]
    ax3 = layout.axs[2]

    if ds_elevation is None:
        ds_elevation = ds

    figs: list[ECKFigure] = []

    for _var, _ax in {
        "synergetic_target_classification": ax1,
        "synergetic_target_classification_medium_resolution": ax2,
        "synergetic_target_classification_low_resolution": ax3,
    }.items():
        _fig = CurtainFigure(ax=_ax)
        _fig = _fig.ecplot(
            ds=ds,
            var=_var,
            height_range=height_range,
            **common_kwargs,  # type: ignore
        )
        figs.append(_fig)

    main_figs: list[ECKFigure] = figs

    subfigs: list[list[ECKFigure]] = [map_figs, main_figs]

    _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
    if logger:
        print_progress(
            f"Plot created (time taken {_dtime}).",
            is_last=True,
            log_msg_prefix=log_msg_prefix,
            logger=logger,
        )

    return QuicklookFigure(layout.fig, subfigs)



"""
**earthcarekit.geo**

Geospatial utilities for handling coordinates, coordinate transformations,
distance calculation, and interpolation.

---
"""

from .convertsions import ecef_to_geo, geo_to_ecef, sequence_geo_to_ecef
from .coordinates import (
    get_central_coords,
    get_central_latitude,
    get_central_longitude,
    get_coords,
)
from .distance import geodesic, get_cumulative_distances, haversine, vincenty
from .grid import create_spherical_grid
from .interpolate import get_coord_between, interpgeo

__all__ = [
    "ecef_to_geo",
    "geo_to_ecef",
    "sequence_geo_to_ecef",
    "get_central_coords",
    "get_central_latitude",
    "get_central_longitude",
    "get_coords",
    "geodesic",
    "get_cumulative_distances",
    "haversine",
    "vincenty",
    "get_coord_between",
    "interpgeo",
    "create_spherical_grid",
]



from logging import Logger

import pandas as pd

from ._exceptions import InvalidInputError
from ._types import TimestampStr, _TimestampInputs


def format_datetime_string(
    datetime_string: str,
    logger: Logger | None = None,
) -> TimestampStr:
    """Formats time string and raises ValueError if unsuccessful."""
    try:
        timestamp = pd.Timestamp(datetime_string)
        if timestamp.tzinfo is None:
            timestamp = timestamp.tz_localize("UTC")
        return timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")
    except ValueError as e:
        msg = f"Given time string '{datetime_string}' is not valid. Here is the original error:"
        if logger:
            logger.exception(f"{msg}\n{e}")
        raise


def parse_time(
    timestamps: list[TimestampStr] | None,
    start_time: TimestampStr | None,
    end_time: TimestampStr | None,
    logger: Logger | None = None,
) -> _TimestampInputs:
    timestamps = (
        []
        if not timestamps
        else [format_datetime_string(t, logger=logger) for t in timestamps]
    )
    start_time = (
        None if not start_time else format_datetime_string(start_time, logger=logger)
    )
    end_time = None if not end_time else format_datetime_string(end_time, logger=logger)

    if isinstance(start_time, TimestampStr) and isinstance(end_time, TimestampStr):
        if start_time > end_time:
            raise InvalidInputError(
                f"Start ({start_time}) time must be greater than end time ({end_time})"
            )
        timestamps = [t for t in timestamps if t < start_time or t > end_time]
    elif isinstance(start_time, TimestampStr):
        timestamps = [t for t in timestamps if t < start_time]
    elif isinstance(end_time, TimestampStr):
        timestamps = [t for t in timestamps if t > end_time]

    return _TimestampInputs(
        timestamps=timestamps,
        time_range=(start_time, end_time),
    )



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap_msi_surface_classification():
    colors = [
        "#808080",
        "#010080",
        "#FF6448",
        "#FFD801",
        "#008001",
        "#81007F",
        "#9470DC",
        "#FEA501",
        "#8C0001",
    ]
    cmap = Cmap(
        name="msi_surface_classification",
        colors=colors,
    ).to_categorical(
        {
            0: "Undefined",
            1: "Water",
            2: "Land",
            3: "Desert",
            4: "Vegetation NDVI",
            5: "Snow XMET",
            6: "Snow NDSI",
            7: "Sea ice XMET",
            8: "Sunglint",
        }
    )
    return cmap


def get_cmap_msi_cloud_phase():
    colors = [
        "#dedede",
        "#1192e8",
        "#93fbff",
        "#123598",
        "#ff2a55",
    ]
    cmap = Cmap(
        name="msi_cloud_phase",
        colors=colors,
    ).to_categorical(
        {
            -127: "Not determined",
            1: "Water",
            2: "Ice",
            3: "S'cooled",
        }
    )
    return cmap


def get_cmap_msi_cloud_mask():
    colors = [
        "#dedede",
        "#123598",
        "#1192e8",
        "#ffaa00",
        "#ff2a55",
    ]
    cmap = Cmap(
        name="msi_cloud_mask",
        colors=colors,
    ).to_categorical(
        {
            -127: "Not determined",
            0: "Clear",
            1: "Prob. clear",
            2: "Prob. cloudy",
            3: "Cloudy",
        }
    )
    return cmap



from dataclasses import dataclass
from logging import Logger

from ._exceptions import InvalidInputError
from ._types import OrbitInt, _OrbitNumbers


def get_validated_orbit_number(
    orbit_number: OrbitInt, logger: Logger | None = None
) -> OrbitInt:
    """Raises InvalidInputError if orbit number is negative or too large"""
    try:
        if orbit_number < 0 or orbit_number > 99999:
            exception_msg = f"{orbit_number} is not a valid orbit number. Valid orbit numbers are positive integers up to 5 digits."
            raise InvalidInputError(exception_msg)
    except InvalidInputError as e:
        if logger:
            logger.exception(e)
        raise
    return orbit_number


def parse_orbit_numbers(
    orbit_numbers: list[int] | None,
    start_orb: int | None,
    end_orb: int | None,
    logger: Logger | None = None,
) -> _OrbitNumbers:
    orb_list: list[OrbitInt] = []
    if isinstance(orbit_numbers, list):
        orb_list = [get_validated_orbit_number(o, logger=logger) for o in orbit_numbers]
    orb_range: tuple[OrbitInt | None, OrbitInt | None] = (
        None if not start_orb else get_validated_orbit_number(start_orb, logger=logger),
        None if not end_orb else get_validated_orbit_number(end_orb, logger=logger),
    )

    if isinstance(orb_range[0], OrbitInt) and isinstance(orb_range[1], OrbitInt):
        orb_list = [o for o in orb_list if o < orb_range[0] or o > orb_range[1]]
    elif isinstance(orb_range[0], OrbitInt):
        orb_list = [o for o in orb_list if o < orb_range[0]]
    elif isinstance(orb_range[1], OrbitInt):
        orb_list = [o for o in orb_list if o > orb_range[1]]

    return _OrbitNumbers(
        orbit_range=orb_range,
        orbit_list=orb_list,
    )



import plotly.colors as pc  # type: ignore

from ..color import Color
from .cmap import Cmap


def _get_plotly_cmaps(color_scale_module: object, smooth: bool) -> dict[str, Cmap]:
    cmaps = {}
    for name in dir(color_scale_module):
        if not name.startswith("_"):
            colors = getattr(color_scale_module, name)
            if isinstance(colors, list):
                if all([isinstance(c, str) for c in colors]):
                    cmap = Cmap(
                        [Color(c).rgba for c in colors], gradient=smooth, name=name
                    )
                    cmaps.update({name: cmap})
    return cmaps


def get_all_plotly_cmaps() -> dict[str, Cmap]:
    plotly_cmaps = {}

    plotly_cmaps.update(_get_plotly_cmaps(pc.sequential, True))
    plotly_cmaps.update(_get_plotly_cmaps(pc.diverging, True))
    plotly_cmaps.update(_get_plotly_cmaps(pc.cyclical, True))
    plotly_cmaps.update(_get_plotly_cmaps(pc.qualitative, False))

    return plotly_cmaps



import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex


def get_color_list():
    """
    Based on color index values from the calipso colormap in the 'ectools' project by Shannon Mason (ECMWF):
    https://bitbucket.org/smason/workspace/projects/EC
    Original under Apache 2.0 license.
    """
    colors = [
        (0.000, Color("#123598")),
        (0.066, Color("#0C3AB8")),
        (0.214, Color("#007FFF")),
        (0.286, Color("#007FFF")),
        (0.302, Color("#00FFAA")),
        (0.313, Color("#007F7F")),
        (0.327, Color("#00AA55")),
        (0.339, Color("#FFFF00")),
        (0.364, Color("#FFFF00")),
        (0.411, Color("#FFE600")),
        (0.446, Color("#FFD400")),
        (0.481, Color("#FFAA00")),
        (0.510, Color("#FF7F00")),
        (0.533, Color("#FF5500")),
        (0.556, Color("#FF0000")),
        (0.580, Color("#FF2A55")),
        (0.597, Color("#FF557F")),
        (0.609, Color("#FF7FAA")),
        (0.621, Color("#464646")),
        (0.632, Color("#5A5A5A")),
        (0.644, Color("#6E6E6E")),
        (0.656, Color("#828282")),
        (0.673, Color("#969696")),
        (0.696, Color("#AAAAAA")),
        (0.720, Color("#B4B4B4")),
        (0.743, Color("#BEBEBE")),
        (0.767, Color("#C8C8C8")),
        (0.790, Color("#D2D2D2")),
        (0.813, Color("#D7D7D7")),
        (0.837, Color("#DCDCDC")),
        (0.860, Color("#E1E1E1")),
        (0.883, Color("#E6E6E6")),
        (0.907, Color("#EBEBEB")),
        (0.932, Color("#F0F0F0")),
        (0.959, Color("#F5F5F5")),
        (0.986, Color("#FFFFFF")),
        (1.000, Color("#FFFFFF")),
    ]
    return colors


def get_cmap() -> Colormap:
    """Creates the Calipso color map."""
    colors = get_color_list()
    cmap = LinearSegmentedColormap.from_list(
        "calipso_smooth", colors, N=256
    ).with_extremes(bad=colors[0][1])
    return cmap



import os
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from xml.etree.ElementTree import Element

import pandas as pd


@dataclass
class HDRFixedHeader:
    file_name: str | None
    file_description: str | None
    notes: str | None
    mission: str | None
    file_class: str | None
    file_type: str | None
    validity_start: pd.Timestamp | None
    validity_stop: pd.Timestamp | None
    file_version: str | None
    system: str | None
    creator: str | None
    creator_version: str | None
    creation_date: str | None


def _safe_get_text(root: Element, path: str) -> str | None:
    element = root.find(path)
    if isinstance(element, Element):
        return element.text
    else:
        return None


def read_hdr_fixed_header(filepath: str) -> HDRFixedHeader:
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"HDR file not found: {filepath}")

    tree = ET.parse(filepath)
    root = tree.getroot()

    validity_start_string = _safe_get_text(
        root, "Fixed_Header/Validity_Period/Validity_Start"
    )
    if isinstance(validity_start_string, str):
        validity_start = pd.Timestamp(validity_start_string.replace("UTC=", ""))
    else:
        validity_start = None

    validity_stop_string = _safe_get_text(
        root, "Fixed_Header/Validity_Period/Validity_Stop"
    )
    if isinstance(validity_stop_string, str):
        validity_stop = pd.Timestamp(validity_stop_string.replace("UTC=", ""))
    else:
        validity_stop = None

    fixed_header_data = HDRFixedHeader(
        file_name=_safe_get_text(root, "Fixed_Header/File_Name"),
        file_description=_safe_get_text(root, "Fixed_Header/File_Description"),
        notes=_safe_get_text(root, "Fixed_Header/Notes"),
        mission=_safe_get_text(root, "Fixed_Header/Mission"),
        file_class=_safe_get_text(root, "Fixed_Header/File_Class"),
        file_type=_safe_get_text(root, "Fixed_Header/File_Type"),
        validity_start=validity_start,
        validity_stop=validity_stop,
        file_version=_safe_get_text(root, "Fixed_Header/File_Version"),
        system=_safe_get_text(root, "Fixed_Header/Source/System"),
        creator=_safe_get_text(root, "Fixed_Header/Source/Creator"),
        creator_version=_safe_get_text(root, "Fixed_Header/Source/Creator_Version"),
        creation_date=_safe_get_text(root, "Fixed_Header/Source/Creation_Date"),
    )
    return fixed_header_data



from dataclasses import dataclass

import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex


def get_cmap():
    name = "labview"
    colors = ["#0043DC", "#00AAC7", "#4FEE25", "#FEF900", "#FF7F00", "#FF1200"]
    positions = [0.0, 0.19929453, 0.3633157, 0.6031746, 0.7707231, 1.0]
    color_dict = list(zip(positions, colors))
    cmap = LinearSegmentedColormap.from_list(name, color_dict).with_extremes(
        under=f"#000000", bad=f"#000000"
    )
    return cmap



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    ELEVATION_VAR,
    TROPOPAUSE_VAR,
    rename_common_dims_and_vars,
    rename_var_info,
)
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_ccd(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens CPR_CD__2A file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="CPR_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        elevation_var="surface_elevation",
        land_flag_var="land_flag",
    )

    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_uncorrected",
        long_name="Uncorrected doppler velocity",
    )
    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_corrected_for_mispointing",
        long_name="Doppler velocity corrected for mispointing",
    )
    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_corrected_for_nubf",
        long_name="Doppler velocity corrected for non-uniform beam filling",
    )
    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_integrated",
        long_name="Integrated doppler velocity",
    )
    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_integrated_error",
        long_name="Integrated doppler velocity error",
    )
    ds = rename_var_info(
        ds=ds,
        var="doppler_velocity_best_estimate",
        long_name="Doppler velocity best est.",
    )

    ds = rename_var_info(
        ds=ds,
        var="sedimentation_velocity_best_estimate",
        long_name="Sedimentation velocity best est.",
    )
    ds = rename_var_info(
        ds=ds,
        var="sedimentation_velocity_best_estimate_error",
        long_name="Sedimentation velocity best est. error",
    )
    ds = rename_var_info(
        ds=ds,
        var="spectrum_width_integrated",
        long_name="Integrated spectrum width",
    )
    ds = rename_var_info(
        ds=ds,
        var="spectrum_width_uncorrected",
        long_name="Uncorrected spectrum width",
    )

    for v in ds.variables:
        if hasattr(ds[v], "units"):
            if ds[v].units == "m s-1":
                ds[v] = ds[v].assign_attrs(units="m/s")

    return ds



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    add_nadir_var,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_actc(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens AC__TC__2B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="JSG_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        elevation_var="elevation",
    )

    ds = rename_var_info(
        ds,
        "synergetic_target_classification",
        long_name="Synergetic target classification (HiRes)",
    )
    ds = rename_var_info(
        ds,
        "synergetic_target_classification_medium_resolution",
        long_name="Synergetic target classification (MedRes)",
    )
    ds = rename_var_info(
        ds,
        "synergetic_target_classification_low_resolution",
        long_name="Synergetic target classification (LowRes)",
    )

    return ds



from typing import Any, Literal, Sequence, TypeAlias

import matplotlib.dates as mdates
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.patches import ConnectionPatch

from ...utils.time import TimestampLike, to_timestamp
from ...utils.typing import Number
from ..color import Color, ColorLike

_NumberTimeOrTuple: TypeAlias = (
    int
    | float
    | TimestampLike
    | tuple[int | float | TimestampLike, int | float | TimestampLike]
)


def _convert_to_matplotlib_axis_value(input: Any | TimestampLike) -> float:
    if isinstance(input, TimestampLike):
        t = to_timestamp(input)
        return mdates.date2num(t)
    return float(input)


def _get_point(
    ax: Axes,
    idx: Literal[1, 2],
    point: _NumberTimeOrTuple,
) -> tuple[float, float]:
    p: tuple[float, float]
    if isinstance(point, Sequence) and not isinstance(point, str):
        p = (
            _convert_to_matplotlib_axis_value(point[0]),
            _convert_to_matplotlib_axis_value(point[1]),
        )
    elif isinstance(point, (Number | TimestampLike)):
        p = (
            _convert_to_matplotlib_axis_value(point),
            ax.get_ylim()[idx - 1],
        )
    else:
        raise TypeError(f"Invalid type {type(point).__name__} for point{idx}")
    return p


def plot_line_between_figures(
    ax1: Axes,
    ax2: Axes,
    point1: _NumberTimeOrTuple,
    point2: _NumberTimeOrTuple | None = None,
    color: ColorLike | None = "ec:red",
    linestyle: str = "dashed",
    linewidth: int | float = 2,
    alpha: int | float = 0.3,
    capstyle: str = "butt",
    zorder: int | float = -20,
    **kwargs,
) -> None:
    """Draws a line connecting a point in one subfigure (ax1) to a point in another (ax2)."""
    p1: tuple[float, float] = _get_point(ax1, 1, point1)
    if point2 is None:
        point2 = point1
    p2: tuple[float, float] = _get_point(ax2, 2, point2)

    con = ConnectionPatch(
        xyA=p1,
        coordsA=ax1.transData,
        xyB=p2,
        coordsB=ax2.transData,
        axesA=ax1,
        axesB=ax2,
        color=Color.from_optional(color),
        linestyle=linestyle,
        linewidth=linewidth,
        alpha=alpha,
        capstyle=capstyle,
        zorder=zorder,
        **kwargs,
    )
    ax1.figure.add_artist(con)



from typing import Literal

import numpy as np
from numpy.typing import ArrayLike

from ...constants import MEAN_EARTH_RADIUS_METERS


def haversine(
    a: ArrayLike,
    b: ArrayLike,
    units: Literal["m", "km"] = "km",
    radius_m: float = MEAN_EARTH_RADIUS_METERS,
):
    """
    Calculates the great-circle (spherical) distance between pairs of latitude/longitude coordinates
    using the haversine formula.

    Args:
        a (ArrayLike): An array-like object of shape (..., 2) containing latitude and longitude
            coordinates in degrees. The last dimension must be 2: (lat, lon).
        b (ArrayLike): An array-like object of the same shape as `a`, containing corresponding
            latitude and longitude coordinates.
        units (Literal["m", "km"], optional): Unit of the output distance. Must be either
            "km" for kilometers or "m" for meters. Defaults to "km".
        radius (float, optional): Radius of the sphere to use for distance calculation.
            Defaults to MEAN_EARTH_RADIUS_METERS (based on WSG 84 ellipsoid: ~6371008.77 meters).
            Note: If `units="km"`, this value is automatically converted to kilometers.

    Returns:
        np.ndarray: Array of great-circle distances between `a` and `b`, in the specified units.
            The shape matches the input shape excluding the last dimension.

    Raises:
        ValueError: If the shapes of `a` and `b` are incompatible or `units` is not one of "m" or "km".

    Examples:
        >>> haversine([51.352757, 12.43392], [38.559, 68.856])
        4537.564747442274
        >>> haversine([0,0], [[0,0], [10,0], [20,0]])
        array([   0.        , 1111.95079735, 2223.90159469])
        >>> haversine([[0,0], [10,0], [20,0]], [[0,0], [10,0], [20,0]])
        array([0., 0., 0.])
    """

    if units not in ["m", "km"]:
        raise ValueError(
            f"{haversine.__name__}() Invalid units : {units}. Use 'm' or 'km' instead."
        )

    radius: float = radius_m
    if units == "km":
        radius = radius / 1000.0

    a = np.array(a)
    b = np.array(b)

    coord_a = np.atleast_2d(a)
    coord_b = np.atleast_2d(b)

    if (coord_a.shape[1] != 2) or (coord_b.shape[1] != 2):
        raise ValueError(
            f"At least one passed array has a wrong shape (a={a.shape}, b={b.shape}). 1d arrays should be of length 2 (i.e. [lat, lon]) and 2d array should have the shape (n, 2)."
        )
    if (coord_a.shape[0] < 1) or (coord_b.shape[0] < 1):
        raise ValueError(
            f"At least one passed array contains no values (a={a.shape}, b={b.shape})."
        )
    if coord_a.shape[0] != coord_b.shape[0]:
        if (coord_a.shape[0] != 1) and (coord_b.shape[0] != 1):
            raise ValueError(
                f"The shapes of passed arrays dont match (a={a.shape}, b={b.shape}). Either both should contain the same number of coordinates or at least one of them should contain a single coordinate."
            )

    coord_a = np.radians(coord_a)
    coord_b = np.radians(coord_b)

    phi_1, lambda_1 = coord_a[:, 0], coord_a[:, 1]
    phi_2, lambda_2 = coord_b[:, 0], coord_b[:, 1]

    hav = lambda theta: (1 - np.cos(theta)) / 2

    h = hav(phi_2 - phi_1) + np.cos(phi_1) * np.cos(phi_2) * hav(lambda_2 - lambda_1)

    d = 2 * radius * np.arcsin(np.sqrt(h))

    if len(a.shape) == 1 and len(b.shape) == 1:
        return d[0]

    return d



from itertools import islice
from logging import Logger
from typing import cast

import pandas as pd

from ..utils._cli import console_exclusive_info
from ._collection_selection import (
    CollectionStr,
    ProductTypeStr,
    get_collection_product_type_dict,
)
from ._eo_collection import EOCollection, get_available_collections
from ._eo_product import get_available_parameters, get_available_products
from ._eo_search_request import EOSearchRequest
from ._parse import get_collection_names_available_to_user, parse_user_type
from ._product_types import get_collection_names_matching_product_availability
from ._types import (
    CollectionStr,
    Entrypoint,
    FrameIDStr,
    OrbitFrameStr,
    OrbitInt,
    ProductTypeVersion,
    ProductVersionStr,
    TimestampStr,
    UserType,
    _BBoxSearch,
    _OrbitFrameInputs,
    _RadiusSearch,
    _SearchInputs,
    _TimestampInputs,
)


def split_list_into_chunks(lst: list, size: int) -> list[list]:
    """Splits a list into chunks or sublists each containing at most N elements"""
    iterator = iter(lst)
    return [list(islice(iterator, size)) for _ in range((len(lst) + size - 1) // size)]


def create_search_request_list(
    entrypoint: Entrypoint,
    search_inputs: _SearchInputs,
    input_user_type: str | None = None,
    candidate_coll_names_user: list[CollectionStr] | None = None,
    logger: Logger | None = None,
) -> list[EOSearchRequest]:

    products: list[ProductTypeVersion] = search_inputs.products
    orbit_and_frames: _OrbitFrameInputs = search_inputs.orbit_and_frames
    timestamps: _TimestampInputs = search_inputs.timestamps
    radius_search: _RadiusSearch = search_inputs.radius_search
    bbox_search: _BBoxSearch = search_inputs.bbox_search
    frame_ids: list[FrameIDStr | None] = cast(
        list[FrameIDStr | None],
        orbit_and_frames.frame_ids,
    )
    if len(frame_ids) == 0:
        frame_ids = [None]

    collections: list[EOCollection] = get_available_collections(
        entrypoint=entrypoint, logger=logger
    )
    collection_product_type_dict: dict[CollectionStr, list[ProductTypeStr]] = (
        get_collection_product_type_dict(collections)
    )

    if isinstance(input_user_type, str):
        user_type: UserType = parse_user_type(input_user_type)
        candidate_coll_names_user = get_collection_names_available_to_user(
            user_type, entrypoint=entrypoint
        )
    elif not isinstance(candidate_coll_names_user, list):
        raise ValueError(f"Missing candidate_coll_names_user")

    if entrypoint == Entrypoint.MAAP:
        candidate_coll_names_user = [
            c if "_MAAP" in c else f"{c}_MAAP" for c in candidate_coll_names_user
        ]
    else:
        candidate_coll_names_user = [
            c if "_MAAP" not in c else c.replace("_MAAP", "")
            for c in candidate_coll_names_user
        ]

    planned_requests: list[EOSearchRequest] = []
    for product in products:
        is_only_timerange: bool = isinstance(
            timestamps.time_range[0], TimestampStr
        ) or isinstance(timestamps.time_range[1], TimestampStr)

        candidate_coll_names_all = get_collection_names_matching_product_availability(
            product,
            collection_product_type_dict,
        )
        candidate_coll_names: list[CollectionStr] = list(
            set(candidate_coll_names_all) & set(candidate_coll_names_user)
        )
        candidate_colls: list[EOCollection] = [
            c for c in collections if c.name in candidate_coll_names
        ]

        for t in timestamps.timestamps:
            new_search_request = EOSearchRequest(
                candidate_collections=candidate_colls,
                product_type=product.type,
                product_version=product.formatted_version,
                radius=radius_search.radius,
                lat=radius_search.lat,
                lon=radius_search.lon,
                bbox=bbox_search.bbox,
                start_time=t,
                end_time=t,
                orbit_number=None,
                start_orbit_number=None,
                end_orbit_number=None,
                frame_id=None,
            )
            planned_requests.append(new_search_request)

        if isinstance(orbit_and_frames.full_orbit_range[0], int) or isinstance(
            orbit_and_frames.full_orbit_range[1], int
        ):
            is_only_timerange = False
            new_search_request = EOSearchRequest(
                candidate_collections=candidate_colls,
                product_type=product.type,
                product_version=product.formatted_version,
                radius=radius_search.radius,
                lat=radius_search.lat,
                lon=radius_search.lon,
                bbox=bbox_search.bbox,
                start_time=timestamps.time_range[0],
                end_time=timestamps.time_range[1],
                orbit_number=None,
                start_orbit_number=orbit_and_frames.full_orbit_range[0],
                end_orbit_number=orbit_and_frames.full_orbit_range[1],
                frame_id=None,
            )
            planned_requests.append(new_search_request)

        if len(orbit_and_frames.full_orbits) > 0:
            max_orbs: int = 50
            full_orbit_chunks = split_list_into_chunks(
                orbit_and_frames.full_orbits, max_orbs
            )
            for chunk in full_orbit_chunks:
                is_only_timerange = False
                new_search_request = EOSearchRequest(
                    candidate_collections=candidate_colls,
                    product_type=product.type,
                    product_version=product.formatted_version,
                    radius=radius_search.radius,
                    lat=radius_search.lat,
                    lon=radius_search.lon,
                    bbox=bbox_search.bbox,
                    start_time=timestamps.time_range[0],
                    end_time=timestamps.time_range[1],
                    orbit_number=chunk,
                    start_orbit_number=None,
                    end_orbit_number=None,
                    frame_id=None,
                )
                planned_requests.append(new_search_request)

        for frame_id, orbit_range in orbit_and_frames.frame_orbit_ranges.items():
            if isinstance(orbit_range[0], int) or isinstance(orbit_range[1], int):
                is_only_timerange = False
                new_search_request = EOSearchRequest(
                    candidate_collections=candidate_colls,
                    product_type=product.type,
                    product_version=product.formatted_version,
                    radius=radius_search.radius,
                    lat=radius_search.lat,
                    lon=radius_search.lon,
                    bbox=bbox_search.bbox,
                    start_time=timestamps.time_range[0],
                    end_time=timestamps.time_range[1],
                    orbit_number=None,
                    start_orbit_number=orbit_range[0],
                    end_orbit_number=orbit_range[1],
                    frame_id=frame_id,
                )
                planned_requests.append(new_search_request)

        for frame_id, orbits in orbit_and_frames.frame_orbits.items():
            if len(orbits) > 0:
                is_only_timerange = False
                new_search_request = EOSearchRequest(
                    candidate_collections=candidate_colls,
                    product_type=product.type,
                    product_version=product.formatted_version,
                    radius=radius_search.radius,
                    lat=radius_search.lat,
                    lon=radius_search.lon,
                    bbox=bbox_search.bbox,
                    start_time=timestamps.time_range[0],
                    end_time=timestamps.time_range[1],
                    orbit_number=orbits,
                    start_orbit_number=None,
                    end_orbit_number=None,
                    frame_id=frame_id,
                )
                planned_requests.append(new_search_request)

        if is_only_timerange:
            for frame_id in frame_ids:
                new_search_request = EOSearchRequest(
                    candidate_collections=candidate_colls,
                    product_type=product.type,
                    product_version=product.formatted_version,
                    radius=radius_search.radius,
                    lat=radius_search.lat,
                    lon=radius_search.lon,
                    bbox=bbox_search.bbox,
                    start_time=timestamps.time_range[0],
                    end_time=timestamps.time_range[1],
                    orbit_number=None,
                    start_orbit_number=None,
                    end_orbit_number=None,
                    frame_id=frame_id,
                )
            planned_requests.append(new_search_request)

    if logger and len(planned_requests) == 0:
        console_exclusive_info()
        logger.warning(
            "There are not enough user inputs to create valid search requests."
        )
        logger.warning(
            "Please ensure that you specify at least individual orbits or timestamps, or alternatively an orbit or time range."
        )

    new_planned_requests: list[EOSearchRequest] = []
    for pr in planned_requests:
        new_planned_requests.extend(pr.split_optimize_requests())

    return new_planned_requests



import os

import xarray as xr

from ._read_nc import read_nc
from ._read_polly import read_polly
from .product import read_product
from .product.file_info import is_earthcare_product


def read_any(input: str | xr.Dataset, **kwargs) -> xr.Dataset:
    """Reads various input types and returns an `xarray.Dataset`.

    This function can read:
        - EarthCARE product files (`.h5`)
        - NetCDF files (`.nc`)
        - Manually processed PollyXT output files (`.txt`)

    Args:
        input (str | xr.Dataset): File path or existing Dataset.
        **kwargs: Additional keyword arguments for specific readers.

    Returns:
        xr.Dataset: Opened dataset.

    Raises:
        ValueError: If the file type is not supported.
        TypeError: If the input type is invalid.
    """
    if isinstance(input, xr.Dataset):
        return input
    elif isinstance(input, str):
        filepath = input

        if is_earthcare_product(filepath=filepath):
            return read_product(filepath, **kwargs)

        filename = os.path.basename(filepath)
        _, ext = os.path.splitext(filename)
        if ext.lower() == ".txt":
            return read_polly(filepath)
        elif ext.lower() == ".nc":
            return read_nc(filepath, **kwargs)

        raise ValueError(f"Reading of file not supported: <{input}>")
    raise TypeError(f"Invalid type '{type(input).__name__}' for input.")



import os

import numpy as np
from xarray import Dataset

from ...constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ...xarray_utils import merge_datasets
from ._trim_to_frame import trim_to_latitude_frame_bounds
from .auxiliary import read_product_xjsg, read_product_xmet
from .file_info import FileType
from .header_group import add_header_and_meta_data
from .level1 import read_product_anom, read_product_cnom, read_product_mrgr
from .level2a import (
    read_product_aaer,
    read_product_aald,
    read_product_acla,
    read_product_acth,
    read_product_aebd,
    read_product_afm,
    read_product_aice,
    read_product_atc,
    read_product_ccd,
    read_product_ccld,
    read_product_cclp,
    read_product_ceco,
    read_product_cfmr,
    read_product_ctc,
    read_product_mcm,
    read_product_mcop,
)
from .level2b import (
    read_product_acmcap,
    read_product_actc,
    read_product_amacd,
    read_product_amcth,
)


def _read_auxiliary_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
    ensure_nans: bool,
    **kwargs,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta, ensure_nans]
    match file_type:
        case FileType.AUX_MET_1D:
            return read_product_xmet(*args, **kwargs)
        case FileType.AUX_JSG_1D:
            return read_product_xjsg(*args, **kwargs)
        case _:
            return None


def _read_level1_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
    ensure_nans: bool,
    **kwargs,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta, ensure_nans]
    match file_type:
        case FileType.ATL_NOM_1B:
            return read_product_anom(*args, **kwargs)
        case FileType.MSI_RGR_1C:
            return read_product_mrgr(*args, **kwargs)
        case FileType.CPR_NOM_1B:
            return read_product_cnom(*args, **kwargs)
        case _:
            return None


def _read_level2a_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
    ensure_nans: bool,
    **kwargs,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta, ensure_nans]
    match file_type:
        case FileType.ATL_AER_2A:
            return read_product_aaer(*args, **kwargs)
        case FileType.ATL_EBD_2A:
            return read_product_aebd(*args, **kwargs)
        case FileType.ATL_TC__2A:
            return read_product_atc(*args, **kwargs)
        case FileType.ATL_CLA_2A:
            return read_product_acla(*args, **kwargs)
        case FileType.ATL_CTH_2A:
            return read_product_acth(*args, **kwargs)
        case FileType.ATL_ALD_2A:
            return read_product_aald(*args, **kwargs)
        case FileType.ATL_ICE_2A:
            return read_product_aice(*args, **kwargs)
        case FileType.ATL_FM__2A:
            return read_product_afm(*args, **kwargs)
        case FileType.MSI_CM__2A:
            return read_product_mcm(*args, **kwargs)
        case FileType.MSI_COP_2A:
            return read_product_mcop(*args, **kwargs)
        case FileType.CPR_TC__2A:
            return read_product_ctc(*args, **kwargs)
        case FileType.CPR_CLD_2A:
            return read_product_ccld(*args, **kwargs)
        case FileType.CPR_FMR_2A:
            return read_product_cfmr(*args, **kwargs)
        case FileType.CPR_CD__2A:
            return read_product_ccd(*args, **kwargs)
        case FileType.CPR_CLP_2A:
            return read_product_cclp(*args, **kwargs)
        case FileType.CPR_ECO_2A:
            return read_product_ceco(*args, **kwargs)
        case _:
            return None


def _read_level2b_product(
    filepath: str,
    file_type: FileType,
    modify: bool,
    header: bool,
    meta: bool,
    ensure_nans: bool,
    **kwargs,
) -> Dataset | None:
    args: list = [filepath, modify, header, meta, ensure_nans]
    match file_type:
        case FileType.AM__ACD_2B:
            return read_product_amacd(*args, **kwargs)
        case FileType.AM__CTH_2B:
            return read_product_amcth(*args, **kwargs)
        case FileType.AC__TC__2B:
            return read_product_actc(*args, **kwargs)
        case FileType.ACM_CAP_2B:
            return read_product_acmcap(*args, **kwargs)
        case _:
            return None


def _read_product(
    filepath: str,
    trim_to_frame: bool = True,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> Dataset:
    """Loads an EarthCARE product file as an `xarray.Dataset`.

    Args:
        filepath (str): Path to the product file.
        trim_to_frame (bool, optional): Whether to trim the dataset to latitude frame bounds. Defaults to True.
        modify (bool): If True, default modifications to the opened dataset will be applied
            (e.g., renaming dimension corresponding to height to "vertical"). Defaults to True.
        header (bool): If True, all header data will be included in the dataframe. Defaults to False.
        meta (bool): If True, select meta data from header (like orbit number and frame ID) will be included in the dataframe. Defaults to True.

    Returns:
        xarray.Dataset: Loaded (and optionally trimmed) dataset.
    """
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"No product found under path: {filepath}")

    file_type = FileType.from_input(filepath)

    args: list = [
        filepath,
        file_type,
        modify,
        False,  # header will be read later
        False,  # meta data will be read later
        ensure_nans,
    ]

    ds = _read_level1_product(*args, **kwargs)
    if not isinstance(ds, Dataset):
        ds = _read_level2a_product(*args, **kwargs)
    if not isinstance(ds, Dataset):
        ds = _read_level2b_product(*args, **kwargs)
    if not isinstance(ds, Dataset):
        ds = _read_auxiliary_product(*args, **kwargs)
    if not isinstance(ds, Dataset):
        raise NotImplementedError(f"Product '{file_type}' not yet supported.")

    if file_type == FileType.AUX_MET_1D:
        trim_to_frame = False

    if modify and trim_to_frame:
        ds = trim_to_latitude_frame_bounds(ds)

    ds = add_header_and_meta_data(filepath, ds, header, meta)

    return ds


def read_product(
    input: str | Dataset,
    trim_to_frame: bool = True,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    in_memory: bool = False,
    **kwargs,
) -> Dataset:
    """Returns an `xarray.Dataset` from a Dataset or EarthCARE file path, optionally loaded into memory.

    Args:
        input (str or xarray.Dataset): Path to a EarthCARE file. If a `xarray.Dataset` is given it will be returned as is.
        trim_to_frame (bool, optional): Whether to trim the dataset to latitude frame bounds. Defaults to True.
        modify (bool, optional): If True, default modifications to the opened dataset will be applied
            (e.g., renaming dimension corresponding to height to "vertical"). Defaults to True.
        header (bool, optional): If True, all header data will be included in the dataframe. Defaults to False.
        meta (bool, optional): If True, select meta data from header (like orbit number and frame ID) will be included in the dataframe. Defaults to True.
        ensure_nans (bool, optional): If True, ensures that _FillValues are set to NaNs even  if encoding of _FillValues or dtype is missing.
            Be aware, if True increases reading time. Defaults to False.
        in_memory (bool, optional): If True, ensures the dataset is fully loaded into memory. Defaults to False.

    Returns:
        xarray.Dataset: The resulting dataset.

    Raises:
        TypeError: If input is not a Dataset or string.
    """
    ds: Dataset
    if isinstance(input, Dataset):
        ds = input
    elif isinstance(input, str):
        kwargs = dict(
            trim_to_frame=trim_to_frame,
            modify=modify,
            header=header,
            meta=meta,
            ensure_nans=ensure_nans,
            **kwargs,
        )
        if in_memory:
            with _read_product(filepath=input, **kwargs) as ds:
                ds = ds.load()
        else:
            ds = _read_product(filepath=input, **kwargs)
    else:
        raise TypeError(
            f"Invalid input type! Expecting a opened EarthCARE dataset (xarray.Dataset) or a path to a EarthCARE product."
        )
    return ds



from ._ecql_atl_aer_2a import ecquicklook_aaer
from ._ecql_atl_cth_2a import ecquicklook_acth
from ._ecql_atl_ebd_2a import ecquicklook_aebd
from ._ecql_atl_tc__2a import ecquicklook_atc
from ._ecql_cpr_cd__2a import ecquicklook_ccd
from ._ecql_cpr_cld_2a import ecquicklook_ccld
from ._ecql_cpr_fmr_2a import ecquicklook_cfmr
from ._ecql_cpr_tc__2a import ecquicklook_ctc



from dataclasses import dataclass
from enum import StrEnum
from typing import TypeAlias

from ..utils._cli._parse._types import (
    CollectionStr,
    FrameIDStr,
    OrbitFrameStr,
    OrbitInt,
    ProductTypeStr,
    ProductTypeVersion,
    ProductVersionStr,
    TimestampStr,
    _BBoxSearch,
    _OrbitAndFrames,
    _OrbitFrameInputs,
    _OrbitNumbers,
    _RadiusSearch,
    _SearchInputs,
    _TimestampInputs,
)
from ._constants import URL_MAAP, URL_OADS


class UserType(StrEnum):
    COMMISSIONING = "commissioning"
    CALVAL = "calval"
    OPEN = "public"


class Entrypoint(StrEnum):
    MAAP = URL_MAAP
    OADS = URL_OADS



from contextlib import nullcontext

import numpy as np
from xarray import Dataset

from ...utils import filter_time
from ...utils.read import read_product
from ...utils.read.product._rebin_xmet_to_vertical_track import (
    rebin_xmet_to_vertical_track,
)
from ...utils.time import TimeRangeLike
from ...utils.typing import DistanceRangeLike
from ..figure import CurtainFigure, ECKFigure, FigureType, SwathFigure
from ..figure.multi_panel import create_multi_figure_layout
from ._quicklook_results import QuicklookFigure


def ecquicklook_deep_convection(
    mrgr: Dataset | str,
    cfmr: Dataset | str,
    ccd: Dataset | str,
    aebd: Dataset | str,
    xmet: Dataset | str | None = None,
    height_range: DistanceRangeLike | None = (-250, 20e3),
    time_range: TimeRangeLike | None = None,
    info_text_loc: str | None = None,
    trim_to_frame: bool = False,
) -> QuicklookFigure:
    """
    Creates a 4 panel quicklook of a storm or deep convective event, displaying:

    - 1st row: RGB image from MSI_RGR_1C
    - 2nd row: Radar reflectivity from CPR_FMR_2A
    - 3rd row: Doppler velocity from CPR_CD__2A
    - 4th row: Total attenuated backscatter from ATL_EBD_2A

    Args:
        ds_mrgr (Dataset): The MSI_RGR_1C product filepath or dataset.
        ds_cfmr (Dataset): The CPR_FMR_2A product filepath or dataset.
        ds_ccd (Dataset): The CPR_CD__2A product filepath or dataset.
        ds_aebd (Dataset): The ATL_EBD_2A product filepath or dataset.
        ds_xmet (Dataset | None, optional): The AUX_MET_1D product filepath or dataset.
            If given, temperature contour lines will be added to the plots. Defaults to None.
        height_range (DistanceRangeLike | None, optional): A height range (i.e., min, max) in meters. Defaults to (-250, 20e3).
        time_range (TimeRangeLike | None, optional): A time range to filter the displayed data. Defaults to None.
        info_text_loc (str | None, optional): The positioning of the orbt, frame and product info text (e.g., "upper right").
            Defaults to None.
        trim_to_frame (bool, optional): Wether the read products should be trimmed to the EarthCARE frame bounds.

    Returns:
        QuicklookFigure: The quicklook object.

    Examples:
        ```python
        import earthcarekit as eck

        df = eck.search_product(
            file_type=["mrgr", "cfmr", "ccd", "aebd", "xmet"],
            orbit_and_frame="07590D",
        ).filter_latest()

        fp_mrgr = df.filter_file_type("mrgr").filepath[-1]
        fp_cfmr = df.filter_file_type("cfmr").filepath[-1]
        fp_ccd = df.filter_file_type("ccd").filepath[-1]
        fp_aebd = df.filter_file_type("aebd").filepath[-1]
        fp_xmet = df.filter_file_type("xmet").filepath[-1]

        ql = eck.ecquicklook_deep_convection(
            mrgr=fp_mrgr,
            cfmr=fp_cfmr,
            ccd=fp_ccd,
            aebd=fp_aebd,
            xmet=fp_xmet,
            time_range=("2025-09-28T18:27:10", None),
            info_text_loc="upper left",
        )
        ```

        ![ecquicklook_deep_convection.png](https://raw.githubusercontent.com/TROPOS-RSD/earthcarekit-docs-assets/refs/heads/main/assets/images/quicklooks/ecquicklook_deep_convection.png)
    """

    def _load_xmet() -> Dataset | None:
        if isinstance(xmet, Dataset):
            return xmet
        elif isinstance(xmet, str):
            return read_product(xmet)
        return None

    with (
        read_product(mrgr, trim_to_frame=trim_to_frame) as ds_mrgr,
        read_product(cfmr, trim_to_frame=trim_to_frame) as ds_cfmr,
        read_product(ccd, trim_to_frame=trim_to_frame) as ds_ccd,
        read_product(aebd, trim_to_frame=trim_to_frame) as ds_aebd,
        nullcontext(_load_xmet()) as ds_xmet,
    ):

        min_time = np.max(
            [
                np.min(ds_mrgr.time.values),
                np.min(ds_cfmr.time.values),
                np.min(ds_ccd.time.values),
                np.min(ds_aebd.time.values),
            ]
        )

        max_time = np.min(
            [
                np.max(ds_mrgr.time.values),
                np.max(ds_cfmr.time.values),
                np.max(ds_ccd.time.values),
                np.max(ds_aebd.time.values),
            ]
        )

        ds_mrgr = filter_time(ds_mrgr, (min_time, max_time))
        ds_cfmr = filter_time(ds_cfmr, (min_time, max_time))
        ds_ccd = filter_time(ds_ccd, (min_time, max_time))
        ds_aebd = filter_time(ds_aebd, (min_time, max_time))

        layout = create_multi_figure_layout(
            rows=[
                FigureType.SWATH,
                FigureType.CURTAIN_75,
                FigureType.CURTAIN_75,
                FigureType.CURTAIN_75,
            ],
            hspace=[0.7, 0.35, 0.35],
        )

        figs: list[ECKFigure] = []

        # 1. Row: MSI RGR RGB
        ax = layout.axs[0]

        f: SwathFigure | CurtainFigure
        f = SwathFigure(ax=ax, ax_style_top="time", ax_style_bottom="geo")
        f = f.ecplot(
            ds=ds_mrgr,
            var="rgb",
            time_range=time_range,
            info_text_loc=info_text_loc,
        )
        f = f.ecplot_coastline(ds_mrgr)
        figs.append(f)

        ds_xmet_vert: Dataset | None = None
        if isinstance(ds_xmet, Dataset):
            ds_xmet_vert = rebin_xmet_to_vertical_track(ds_xmet, ds_aebd)
            ds_xmet_vert = filter_time(ds_xmet_vert, time_range)

        # 2. Row CPR FMR reflectivity (Range -40 - 20 dBz)
        ax = layout.axs[1]
        f = CurtainFigure(
            ax=ax,
            ax_style_top="none",
            ax_style_bottom="distance_notitle",
        )
        f = f.ecplot(
            ds=ds_cfmr,
            var="reflectivity_corrected",
            height_range=height_range,
            time_range=time_range,
            value_range=(-40, 20),
            info_text_loc=info_text_loc,
        )
        f = f.ecplot_elevation(ds_cfmr)
        f = f.ecplot_tropopause(ds_aebd)
        if isinstance(ds_xmet_vert, Dataset):
            f = f.ecplot_temperature(ds_xmet_vert)
        figs.append(f)

        # 3. Row CPR-CD Doppler Velocity best estimate (Range -5 -5 m/s)
        ax = layout.axs[2]
        f = CurtainFigure(
            ax=ax,
            ax_style_top="none",
            ax_style_bottom="distance_notitle",
        )
        f = f.ecplot(
            ds=ds_ccd,
            var="doppler_velocity_best_estimate",
            height_range=height_range,
            time_range=time_range,
            value_range=(-5, 5),
            info_text_loc=info_text_loc,
        )
        f = f.ecplot_elevation(ds_cfmr)
        f = f.ecplot_tropopause(ds_aebd)
        if isinstance(ds_xmet_vert, Dataset):
            f = f.ecplot_temperature(ds_xmet_vert)
        figs.append(f)

        # 4. Row ATL-EBD total attenuated mie backscatter
        ax = layout.axs[3]
        f = CurtainFigure(
            ax=ax,
            ax_style_top="none",
            ax_style_bottom="distance",
        )
        f = f.ecplot(
            ds=ds_aebd,
            var="mie_total_attenuated_backscatter_355nm",
            height_range=height_range,
            time_range=time_range,
            info_text_loc=info_text_loc,
        )
        f = f.ecplot_elevation(ds_cfmr)
        f = f.ecplot_tropopause(ds_aebd)
        if isinstance(ds_xmet_vert, Dataset):
            f = f.ecplot_temperature(ds_xmet_vert, colors="white")
        figs.append(f)

        return QuicklookFigure(
            fig=layout.fig,
            subfigs=[figs],
        )



from dataclasses import dataclass
from enum import StrEnum
from typing import TypeAlias

OrbitFrameStr: TypeAlias = str
FrameIDStr: TypeAlias = str
OrbitInt: TypeAlias = int
TimestampStr: TypeAlias = str
CollectionStr: TypeAlias = str
ProductTypeStr: TypeAlias = str
ProductVersionStr: TypeAlias = str


@dataclass
class ProductTypeVersion:
    type: str
    version: str

    @property
    def formatted_version(self) -> str | None:
        if self.version == "latest":
            return None
        return self.version


@dataclass
class _TimestampInputs:
    timestamps: list[TimestampStr]
    time_range: tuple[TimestampStr | None, TimestampStr | None]


@dataclass
class _OrbitNumbers:
    orbit_range: tuple[OrbitInt | None, OrbitInt | None]
    orbit_list: list[OrbitInt]


@dataclass
class _OrbitAndFrames:
    full_orbit_range: tuple[OrbitInt | None, OrbitInt | None]
    full_orbit_list: list[OrbitInt]
    frame_orbits: dict[OrbitFrameStr, list[OrbitInt]]


@dataclass
class _RadiusSearch:
    radius: str | None
    lat: str | None
    lon: str | None


@dataclass
class _BBoxSearch:
    bbox: str | None


@dataclass
class _OrbitFrameInputs:
    full_orbits: list[OrbitInt]
    full_orbit_range: tuple[OrbitInt | None, OrbitInt | None]
    frame_orbits: dict[FrameIDStr, list[OrbitInt]]
    frame_orbit_ranges: dict[FrameIDStr, tuple[OrbitInt | None, OrbitInt | None]]
    frame_ids: list[FrameIDStr]


@dataclass
class _SearchInputs:
    products: list[ProductTypeVersion]
    orbit_and_frames: _OrbitFrameInputs
    timestamps: _TimestampInputs
    radius_search: _RadiusSearch
    bbox_search: _BBoxSearch



import re
import textwrap
from typing import Literal, TypeAlias

from ...utils.debug import get_calling_function_name

AxisInput: TypeAlias = Literal["x", "y", 0, 1]


def validate_axis_input(axis: AxisInput) -> Literal["x", "y"]:
    _axis: Literal["x", "y"]
    if axis in [0, "0", "x"]:
        _axis = "x"
    elif axis in [1, "1", "y"]:
        _axis = "y"
    else:
        raise ValueError(
            f"{get_calling_function_name(2)}() Invalid values given for `axis`: '{axis}' (expecting 'x', 'y' or respectively 0, 1)"
        )
    return _axis



from typing import Any

from xarray import Dataset


def insert_var(
    ds: Dataset,
    var: str,
    data: Any,
    index: int | None = None,
    before_var: str | None = None,
    after_var: str | None = None,
) -> Dataset:
    """
    Inserts a new variable in a `xarray.Dataset` before or after a given variable or at a given index.

    Args:
        ds (Dataset):
            The original dataset to which the variable will be added.
        var (str):
            Name of the new variable to be added.
        data (Any):
            Data stored in the new variable.
        index (int | None, optional):
            Index at which the new variable will be added. Will be ignored when either `before_var` or
            `after_var` are given and valid. Defaults to None.
        before_var (str | None, optional):
            Name of the variable before which the new variable should be inserted. Defaults to None.
        after_var (str | None, optional):
            Name of the variable after which the new variable should be inserted. Will be ignored
            when `before_var` is given and valid. Defaults to None.

    Returns:
        Dataset: The original dataset with the new variable inserted.
    """
    if var in ds.data_vars:
        ds = ds.drop_vars(var)

    if (
        isinstance(index, int)
        or isinstance(before_var, str)
        or isinstance(after_var, str)
    ):
        vars = list(ds.data_vars)

        if isinstance(before_var, str) and before_var in vars:
            index = vars.index(before_var)
        elif isinstance(after_var, str) and after_var in vars:
            index = vars.index(after_var) + 1
        elif not isinstance(index, int):
            index = len(vars)

        vars.insert(index, var)

        ds[var] = data
        ds = ds[vars]
    else:
        ds[var] = data

    return ds



import matplotlib.patheffects as pe
from matplotlib.text import Text

from .color import Color, ColorLike


def add_shade_to_text(
    t: Text,
    alpha: float = 0.8,
    linewidth: float = 3,
    color: Color | ColorLike | None = None,
) -> Text:
    """Applies a shaded stroke effect around a Matplotlib text object.

    Args:
        t (Text): Matplotlib text object to apply the effect to.
        alpha (float, optional): Opacity of the stroke. Defaults to 0.8.
        linewidth (float, optional): Width of the stroke line. Defaults to 3.
        color (Color | ColorLike, optional): Color of the stroke. Defaults to "white".

    Returns:
        Text: The text object with the stroke effect applied.
    """

    if color is None:
        c = Color.from_optional(t.get_color())  # type: ignore
        color = c.get_best_bw_contrast_color()  # type: ignore
    else:
        color = Color.from_optional(color)

    t.set_path_effects(
        [pe.withStroke(linewidth=linewidth, foreground=color, alpha=alpha)]
    )
    return t



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap(
    alpha_clear: float = 1.0,
    alpha_missing: float = 1.0,
    alpha_attenuated: float = 1.0,
    alpha: float = 1.0,
):
    definitions = {
        -3: "Surface",
        -2: "No retrievals",
        -1: "Attenuated",
        0: "Clear sky",
        1: "Likely clear sky",
        2: "Likely clear sky",
        3: "Likely clear sky",
        4: "Likely clear sky",
        5: "Low altitude aerosols",
        6: "Aerosol/thin cloud",
        7: "Aerosol/thin cloud",
        8: "Thick aerosol/cloud",
        9: "Thick aerosol/cloud",
        10: "Thick cloud",
    }

    colors = [
        Color("#000000").set_alpha(alpha_missing),  # Surface
        Color("#FFFFFF").set_alpha(alpha_missing),  # No retrievals
        Color("#283E91").set_alpha(alpha_attenuated),  # Attenuated
        Color("#31BAF0").set_alpha(alpha_clear),  # Clear sky
        Color("#76CFDF").set_alpha(alpha_clear),  # Likely clear sky
        Color("#88D4E4").set_alpha(alpha_clear),  # Likely clear sky
        Color("#98DAE8").set_alpha(alpha_clear),  # Likely clear sky
        Color("#A4DDF0").set_alpha(alpha_clear),  # Likely clear sky
        Color("#BCB8B8").set_alpha(alpha),  # Low altitude aerosols
        Color("#F0DA13").set_alpha(alpha),  # Aerosol/thin cloud
        Color("#F3AA19").set_alpha(alpha),  # Aerosol/thin cloud
        Color("#CD7320").set_alpha(alpha),  # Thick aerosol/cloud
        Color("#F32320").set_alpha(alpha),  # Thick aerosol/cloud
        Color("#9F1D24").set_alpha(alpha),  # Thick cloud
    ]
    cmap = Cmap(colors=colors, name="featuremask").to_categorical(definitions)
    return cmap



import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from numpy.typing import NDArray

from ....utils.time import (
    TimestampLike,
    check_if_same_timestamp,
    format_time_range_text,
    get_time_range,
    lookup_value_by_timestamp,
    to_timestamp,
)


def format_time_ticks(
    ax: Axes,
    time: NDArray,
    lon: NDArray | None = None,
    tmin: TimestampLike | None = None,
    tmax: TimestampLike | None = None,
    show_lst: bool = True,
    show_utc: bool = True,
    show_major: bool = True,
    show_minor: bool = True,
    show_title: bool = True,
    show_tick_labels: bool = True,
    major_frequency: str | None = None,
    minor_frequency: str | None = None,
    xlabel_prefix: str = "",
    max_ticks: int = 12,
    fontsize_ticks: str = "small",
    fontsize_label: str = "small",
):
    if not isinstance(ax, Axes):
        raise TypeError(
            f"{format_time_ticks.__name__}() for `ax` expected type '{Axes.__name__}' but got '{type(ax).__name__}' instead"
        )

    utc_time = xr.DataArray(time)
    if not isinstance(tmin, TimestampLike):
        tmin = utc_time.values[0]
    if not isinstance(tmax, TimestampLike):
        tmax = utc_time.values[-1]

    start_time = to_timestamp(tmin)
    end_time = to_timestamp(tmax)

    duration = end_time - start_time

    divisors = {
        ("1s", "250ms"): 1,  # 1 sec
        ("2s", "1s"): 2,  # 2 sec
        ("3s", "1s"): 3,  # 3 sec
        ("4s", "1s"): 4,  # 4 sec
        ("5s", "1s"): 5,  # 5 sec
        ("10s", "2s"): 10,  # 10 sec
        ("15s", "3s"): 15,  # 15 sec
        ("20s", "5s"): 20,  # 20 sec
        ("30s", "10s"): 30,  # 30 sec
        ("1min", "15s"): 60 * 1,  # 1 min
        ("2min", "30s"): 60 * 2,  # 2 min
        ("3min", "1min"): 60 * 3,  # 3 min
        ("4min", "1min"): 60 * 4,  # 4 min
        ("5min", "1min"): 60 * 5,  # 5 min
        ("10min", "2min"): 60 * 10,  # 10 min
        ("15min", "3min"): 60 * 15,  # 15 min
        ("20min", "5min"): 60 * 20,  # 20 min
        ("30min", "10min"): 60 * 30,  # 30 min
        ("1h", "15min"): 60 * 60 * 1,  # 1 hour
        ("2h", "30min"): 60 * 60 * 2,  # 2 hours
        ("3h", "1h"): 60 * 60 * 3,  # 3 hours
        ("4h", "1h"): 60 * 60 * 4,  # 4 hours
        ("6h", "2h"): 60 * 60 * 6,  # 6 hours
        ("12h", "3h"): 60 * 60 * 12,  # 12 hours
        ("1D", "6h"): 60 * 60 * 24 * 1,  # 1 day
        ("2D", "12h"): 60 * 60 * 24 * 2,  # 2 days
        ("3D", "1D"): 60 * 60 * 24 * 3,  # 3 days
        ("4D", "1D"): 60 * 60 * 24 * 4,  # 4 days
        ("5D", "1D"): 60 * 60 * 24 * 5,  # 5 days
        ("1W", "1D"): 60 * 60 * 24 * 7,  # 7 days
        ("2W", "1W"): 60 * 60 * 24 * 14,  # 14 days
        ("MS", "SMS"): 60 * 60 * 24 * 30,  # 1 month
        ("QS", "MS"): 60 * 60 * 24 * 30 * 2,  # 2 months
        ("QS", "MS"): 60 * 60 * 24 * 30 * 6,  # 6 months
        ("1YS", "QS"): 60 * 60 * 24 * 356,  # 1 year
        ("2YS", "1YS"): 60 * 60 * 24 * 356 * 2,  # 2 years
    }

    for key, value in divisors.items():
        if (duration.days * 24 * 60 * 60 + duration.seconds) / value <= max_ticks:
            major_frequency = key[0]
            minor_frequency = key[1]
            break

    format_time = lambda t: t.strftime("%H:%M:%S")
    if "s" in major_frequency:
        format_time = lambda t: f"{t:%H:%M:%S}"
    elif "min" in major_frequency:
        format_time = lambda t: f"{t:%H:%M}"
    elif "h" in major_frequency:
        format_time = lambda t: f"{t:%m-%d %H}"
        if f"{start_time:%Y-%m-%d}" == f"{end_time:%Y-%m-%d}":
            format_time = lambda t: f"{t:%H:%M}"
    elif "D" in major_frequency or "W" in major_frequency:
        format_time = lambda t: f"{t:%Y-%m-%d}"
    elif "MS" in major_frequency or "SMS" in major_frequency or "QS" in major_frequency:
        format_time = lambda t: f"{t:%Y-%m-%d}"
    elif "Y" in major_frequency:
        format_time = lambda t: f"{t:%Y-%m-%d}"

    utc_ticks = get_time_range(start_time, end_time, freq=major_frequency)
    if len(utc_ticks) == 0:
        raise ValueError(f"invalid time range: ({start_time}, {end_time})")
    lst_time = None
    if lon is not None:
        lon = xr.DataArray(lon)
        lons = [
            lookup_value_by_timestamp(_t, utc_time.values, lon.to_numpy())
            for _t in utc_ticks
        ]
        lst_time = xr.DataArray(utc_ticks) + [
            pd.Timedelta(l / 15, "h").to_numpy() for l in lons
        ]
    utc_tick_labels = [format_time(t) for t in utc_ticks]

    time_idx = [np.argmin(np.abs(pd.DatetimeIndex(utc_ticks) - t)) for t in utc_ticks]
    if lst_time is not None:
        lst_ticks = pd.DatetimeIndex(lst_time[time_idx])
        lst_tick_labels = [format_time(t) for t in lst_ticks]
    else:
        show_lst = False

    tick_labels = ["" for utc in utc_tick_labels]
    if show_utc and show_lst:
        tick_labels = [
            f"{utc}\n{lst}" for utc, lst in zip(utc_tick_labels, lst_tick_labels)
        ]
        last_tick_label = tick_labels[-1].split("\n")
        tick_labels[-1] = (
            f"          {last_tick_label[0]} (UTC)\n         {last_tick_label[1]} (LST)"
        )
    elif show_utc:
        tick_labels = [f"{utc}" for utc in utc_tick_labels]
        last_tick_label = tick_labels[-1]
        tick_labels[-1] = f"{last_tick_label}"
    elif show_lst:
        tick_labels = [f"{lst}" for lst in lst_tick_labels]
        last_tick_label = tick_labels[-1]
        tick_labels[-1] = f"{last_tick_label}"

    ax.set_xticks([])
    ax.set_xticks([], minor=True)
    ax.set_xlabel("")

    if show_major:
        ax.set_xticks(
            utc_ticks,
            labels=tick_labels if show_tick_labels else ["" for x in tick_labels],
            fontsize=fontsize_ticks,
        )

    if show_minor:
        utc_ticks_minor = get_time_range(start_time, end_time, freq=minor_frequency)
        ax.set_xticks(utc_ticks_minor, minor=True)

    if show_title and show_tick_labels:
        label = f"{xlabel_prefix}{pd.to_datetime(utc_ticks[0]):%d %b %Y}"
        if not check_if_same_timestamp(start_time, end_time).is_same:
            label = f"{xlabel_prefix}{format_time_range_text(start_time, end_time, show_date=True, show_time=False)}"
        if show_utc and not show_lst:
            label = label + " (UTC)"
        elif not show_utc and show_lst:
            label = label + " (LST)"
        ax.set_xlabel(label, loc="center", fontsize=fontsize_label)



from ._concat import concat_datasets, read_products
from ._generic import read_product
from ._header_file import read_hdr_fixed_header
from ._rebin_msi_to_jsg import rebin_msi_to_jsg
from ._rebin_xmet_to_vertical_track import rebin_xmet_to_vertical_track
from ._search import search_product
from ._trim_to_frame import trim_to_latitude_frame_bounds
from .file_info import get_product_infos
from .header_group import extract_basic_meta_data_from_header, read_header_data
from .level1.msi_rgr_1c import _add_rgb as update_rgb_of_mrgr
from .level2a.msi_cop_2a import add_isccp_cloud_type
from .science_group import read_science_data



from dataclasses import dataclass
from logging import Logger

import numpy as np
import pandas as pd
from requests.exceptions import HTTPError

from ..utils._cli import get_counter_message
from ..utils.time import time_to_iso, to_timestamp
from ._eo_collection import EOCollection
from ._eo_product import (
    EOProduct,
    get_available_products,
    remove_duplicates_keeping_latest,
)


@dataclass
class EOSearchRequest:
    """This class contains all data required as input for the URL template of the OpenSearch API request to EO-CAT."""

    candidate_collections: list[EOCollection]
    product_type: str | None = None
    product_version: str | None = None
    radius: str | None = None
    lat: str | None = None
    lon: str | None = None
    bbox: str | None = None
    start_time: str | None = None
    end_time: str | None = None
    orbit_number: list[int] | None = None
    start_orbit_number: int | None = None
    end_orbit_number: int | None = None
    frame_id: str | None = None
    limit: int = 2000
    orbit_direction: str | None = None
    instrument: str | None = None
    product_id: str | None = None
    illumination_elevation_angle: str | None = None

    def copy(self) -> "EOSearchRequest":
        return EOSearchRequest(
            candidate_collections=self.candidate_collections,
            product_type=self.product_type,
            product_version=self.product_version,
            radius=self.radius,
            lat=self.lat,
            lon=self.lon,
            bbox=self.bbox,
            start_time=self.start_time,
            end_time=self.end_time,
            orbit_number=self.orbit_number,
            start_orbit_number=self.start_orbit_number,
            end_orbit_number=self.end_orbit_number,
            frame_id=self.frame_id,
            limit=self.limit,
            orbit_direction=self.orbit_direction,
            instrument=self.instrument,
            product_id=self.product_id,
            illumination_elevation_angle=self.illumination_elevation_angle,
        )

    def split_optimize_requests(self) -> list["EOSearchRequest"]:
        new_requests: list[EOSearchRequest] = []

        num_frames: int = 1 if self.frame_id else 8

        num_orbits: int = 0
        if isinstance(self.end_orbit_number, int):
            _so = 0 if not self.start_orbit_number else self.start_orbit_number
            if _so == self.end_orbit_number:
                new_r = self.copy()
                new_r.start_orbit_number = _so
                new_r.end_orbit_number = _so
                _last_eo = new_r.end_orbit_number
                new_requests.append(new_r)
            else:
                num_orbits = self.end_orbit_number - _so
                num_files = num_orbits * num_frames
                num_new_requests = int(np.ceil(num_files / self.limit))
                _last_eo = 0
                for i in range(num_new_requests):
                    new_r = self.copy()
                    new_r.start_orbit_number = max(
                        _last_eo, int(_so + np.floor(num_orbits / num_new_requests) * i)
                    )
                    new_r.end_orbit_number = int(
                        _so + np.ceil(num_orbits / num_new_requests) * (i + 1)
                    )
                    _last_eo = new_r.end_orbit_number
                    new_requests.append(new_r)
        elif isinstance(self.start_orbit_number, int):
            _so = self.start_orbit_number
            _t_ref = pd.Timestamp("2025-08-01")
            _o_ref = 6675
            _t_now = pd.Timestamp.now()
            _t_delta = _t_now - _t_ref
            _days = _t_delta.total_seconds() / (60 * 60 * 24)
            _eo = int(_o_ref + (_days * 16))
            num_orbits = _eo - _so

            num_files = num_orbits * num_frames
            num_new_requests = int(np.ceil(num_files / self.limit))
            _last_eo = 0
            for i in range(num_new_requests):
                new_r = self.copy()
                new_r.start_orbit_number = max(
                    _last_eo, int(_so + np.floor(num_orbits / num_new_requests) * i)
                )
                new_r.end_orbit_number = int(
                    _so + np.floor(num_orbits / num_new_requests) * (i + 1)
                )
                _last_eo = new_r.end_orbit_number
                new_requests.append(new_r)
        elif self.orbit_number:
            num_orbits = len(self.orbit_number)

            num_files = num_orbits * num_frames
            num_new_requests = int(np.ceil(num_files / self.limit))
            for i in range(num_new_requests):
                new_r = self.copy()
                _idx1 = int(np.floor(num_orbits / num_new_requests) * i)
                _idx2 = int(np.floor(num_orbits / num_new_requests) * (i + 1))
                new_r.orbit_number = self.orbit_number[_idx1:_idx2]
                new_requests.append(new_r)
        elif self.end_time:
            _et = to_timestamp(self.end_time)
            _st = (
                to_timestamp("2024-07-31")
                if not self.start_time
                else to_timestamp(self.start_time)
            )
            _t_delta = _et - _st
            _days = _t_delta.total_seconds() / (60 * 60 * 24)
            num_orbits = max(1, int(_days * 16.0))

            num_files = num_orbits * num_frames
            num_new_requests = int(np.ceil(num_files / self.limit))
            for i in range(num_new_requests):
                new_r = self.copy()
                new_r.start_time = time_to_iso(_st + (_t_delta / num_new_requests) * i)
                new_r.end_time = time_to_iso(
                    _st + (_t_delta / num_new_requests) * (i + 1)
                )
                new_requests.append(new_r)
        elif self.start_time:
            _et = pd.Timestamp.now()
            _st = to_timestamp(self.start_time)
            _t_delta = _et - _st
            _days = _t_delta.total_seconds() / (60 * 60 * 24)
            num_orbits = max(1, int(_days * 16.0))

            num_files = num_orbits * num_frames
            num_new_requests = int(np.ceil(num_files / self.limit))
            for i in range(num_new_requests):
                new_r = self.copy()
                new_r.start_time = time_to_iso(_st + (_t_delta / num_new_requests) * i)
                new_r.end_time = time_to_iso(
                    _st + (_t_delta / num_new_requests) * (i + 1)
                )
                new_requests.append(new_r)
        else:
            new_requests = [self.copy()]

        num_files = num_orbits * num_frames

        return new_requests

    @property
    def stac_parameters(self) -> dict[str, str]:
        params: dict[str, str] = {}
        if isinstance(self.limit, int):
            params["limit"] = str(self.limit + 1)
        if isinstance(self.orbit_direction, str):
            params["orbitDirection"] = self.orbit_direction
        if isinstance(self.instrument, str):
            params["instrument"] = self.instrument
        if isinstance(self.product_id, str):
            params["uid"] = self.product_id
        if isinstance(self.illumination_elevation_angle, str):
            params["illuminationElevationAngle"] = self.illumination_elevation_angle
        # Product
        if isinstance(self.product_type, str):
            params["productType"] = self.product_type
        if isinstance(self.product_version, str):
            params["productVersion"] = self.product_version.lower()
        # Geo search
        if isinstance(self.radius, str):
            params["radius"] = self.radius
        if isinstance(self.lat, str):
            params["lat"] = self.lat
        if isinstance(self.lon, str):
            params["lon"] = self.lon
        if isinstance(self.bbox, str):
            params["bbox"] = self.bbox
        # Orbit number and frame ID
        if isinstance(self.orbit_number, list):
            orb_num_list_str = ",".join([str(int(o)) for o in self.orbit_number])
            params["orbitNumber"] = "{" + orb_num_list_str + "}"
        elif isinstance(self.start_orbit_number, int) and isinstance(
            self.end_orbit_number, int
        ):
            params["orbitNumber"] = (
                f"[{str(self.start_orbit_number)},{str(self.end_orbit_number)}]"
            )
        elif isinstance(self.start_orbit_number, int):
            params["orbitNumber"] = f"[{str(self.start_orbit_number)},99999]"
        elif isinstance(self.end_orbit_number, int):
            params["orbitNumber"] = f"[0,{str(self.end_orbit_number)}]"
        if isinstance(self.frame_id, str):
            params["frame"] = self.frame_id
        # Time search
        if isinstance(self.start_time, str) and isinstance(self.end_time, str):
            params["datetime"] = f"{self.start_time}/{self.end_time}"
        elif isinstance(self.start_time, str):
            params["datetime"] = f"{self.start_time}/"
        elif isinstance(self.end_time, str):
            params["datetime"] = f"/{self.end_time}"

        return params

    def run(
        self,
        logger: Logger | None = None,
        total_count: int | None = None,
        counter: int | None = None,
        download_only_h5: bool = False,
    ) -> list[EOProduct]:
        count_msg, _ = get_counter_message(counter=counter, total_count=total_count)

        if logger:
            logger.info(f"*{count_msg} {self.low_detail_summary}")
            logger.debug(f" {count_msg} {self}")

        if len(self.candidate_collections) == 0:
            if logger:
                logger.warning(
                    f" {count_msg} No collection was selected. Please make sure that you have added the appropriate collections for this product in the configuration file and that you are allowed to access to them."
                )
                return []

        _available_products: list[EOProduct] = []
        for cc in sorted(self.candidate_collections):

            try:
                _available_products = get_available_products(
                    cc,
                    params=self.stac_parameters,
                    logger=logger,
                    download_only_h5=download_only_h5,
                )
            except HTTPError as e:
                if logger:
                    logger.exception(e)
                    if (
                        self.stac_parameters.get("productType") == "AUX_MET_1D"
                        and self.stac_parameters.get("radius") is not None
                    ):
                        logger.error("Radius search is not supported for AUX_MET_1D.")

            _available_products = remove_duplicates_keeping_latest(_available_products)

            if len(_available_products) > 0:
                if logger:
                    logger.info(
                        f" {count_msg} Files found in collection '{cc.name}': {len(_available_products)}"
                    )
                    if len(_available_products) >= self.limit:
                        logger.warning(
                            f"The number of results equals the limit of results per search request ({self.limit}). Please divide your requests into smaller, manageable chunks (e.g., by selecting a shorter time range)."
                        )
                break

        return _available_products

    @property
    def low_detail_summary(self):
        msg = f"Search request: {self.product_type}"
        if self.product_version:
            msg = f"{msg}:{self.product_version}"
        if self.start_time and self.end_time:
            if self.start_time == self.end_time:
                msg = f"{msg}, time={self.start_time}"
            else:
                st_msg = self.start_time if self.start_time else "..."
                et_msg = self.end_time if self.end_time else "..."
                msg = f"{msg}, time_range=({st_msg},{et_msg})"
        if self.radius and self.lat and self.lon:
            msg = f"{msg}, radius=({self.radius}m, {self.lat}N, {self.lon}E)"
        if self.bbox:
            msg = f"{msg}, bbox={self.bbox}"
        if self.frame_id:
            msg = f"{msg}, frame={self.frame_id}"
        if self.start_orbit_number or self.end_orbit_number:
            so_msg = self.start_orbit_number if self.start_orbit_number else "..."
            eo_msg = self.end_orbit_number if self.end_orbit_number else "..."
            msg = f"{msg}, orbit_range=({so_msg},{eo_msg})"
        if self.orbit_number:
            o_msg = f"{self.orbit_number}"
            if len(o_msg.split(",")) > 6:
                o_msg = (
                    ",".join(o_msg.split(",")[0:2])
                    + f",... {len(o_msg.split(',')) - 4} more orbits ...,"
                    + ",".join(o_msg.split(",")[-2:])
                )
            msg = f"{msg}, orbits={o_msg.replace(',', ', ')}"
        return msg



import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    EXT_LABEL,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    add_nadir_var,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_acmcap(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens ACM_CAP_2B file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        vertical_dim="JSG_height",
        track_lat_var="latitude",
        track_lon_var="longitude",
        height_var="height",
        time_var="time",
        elevation_var="elevation",
        tropopause_var="tropopause_height",
    )

    "ice_water_content",
    "ice_effective_radius",
    "rain_water_content",
    "rain_median_volume_diameter",
    "liquid_water_content",
    "liquid_effective_radius",
    "aerosol_extinction",

    ds = rename_var_info(
        ds=ds,
        var="ice_water_content",
        long_name="Ice water content",
        units="kg m$^{-3}$",
    )
    ds = rename_var_info(
        ds=ds,
        var="ice_effective_radius",
        long_name="Ice effective radius",
    )
    ds = rename_var_info(
        ds=ds,
        var="rain_water_content",
        long_name="Rain water content",
        units="kg m$^{-3}$",
    )
    ds = rename_var_info(
        ds=ds,
        var="rain_median_volume_diameter",
        long_name="Rain median volume diameter",
    )
    ds = rename_var_info(
        ds=ds,
        var="liquid_water_content",
        long_name="Liquid water content",
        units="kg m$^{-3}$",
    )
    ds = rename_var_info(
        ds=ds,
        var="liquid_effective_radius",
        long_name="Liquid effective radius",
    )
    ds = rename_var_info(
        ds=ds,
        var="aerosol_extinction",
        long_name=EXT_LABEL,
        units="m$^{-1}$",
    )

    return ds



import inspect


def get_calling_function_name(level: int = 1):
    """
    Returns the name of the function at a specified `level` in the call stack.

    Parameters:
        level (int):
            The number of levels up the call stack to look for the calling function.
            A level of 1 refers to the immediate caller, 2 to the caller of the caller, etc.

    Returns:
        function_name (str):
            The name of the function at the given `level` in the call hierarchy.
    """
    stack = inspect.stack()
    if len(stack) >= level:
        return stack[level].function
    return ""



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap

_COLORS = [
    "#E6E6E6",
    "#FFFFFF",
    "#0B4D79",
    "#17BECF",
    "#93FBFF",
    "#2CA02C",
    "#FFFF9A",
    "#E5AE38",
    "#FF7E0E",
    "#D62628",
    "#5D003F",
]


def get_cmap():
    colors = _COLORS
    definitions = {
        -127: "No data",
        0: "Clear",
        1: "Cumulus",
        2: "Altocumulus",
        3: "Cirrus",
        4: "Stratocumulus",
        5: "Altostratus",
        6: "Cirrostratus",
        7: "Stratus",
        8: "Nimbostratus",
        9: "Deep convection",
    }

    cmap = (
        Cmap(
            colors=colors,
            name="msi_cloud_type",
        )
        .to_categorical(definitions)
        .with_extremes(under="#FFFFFF00", over="#FFFFFF00")
    )
    return cmap


def get_cmap_with_short_labels():
    colors = _COLORS
    definitions = {
        -127: "No data",
        0: "Clear",
        1: "Cu",
        2: "Ac",
        3: "Ci",
        4: "Sc",
        5: "As",
        6: "Cs",
        7: "St",
        8: "Ns",
        9: "Dcv",
    }

    cmap = (
        Cmap(
            colors=colors,
            name="msi_cloud_type_short_labels",
        )
        .to_categorical(definitions)
        .with_extremes(under="#FFFFFF00", over="#FFFFFF00")
    )
    return cmap
